'''
This module implements the FeatureTable object, which is mostly a 
wrapper around a pandas dataframe. This also includes methods to 
QAQC and batch correct the feature table. 
'''

import os
import sys
import json
from functools import partial
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import pandas as pd
import intervaltree
from combat.pycombat import pycombat
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from matplotlib.patches import Patch
from . import utils

class FeatureTable:
    """
    A feature table is a data frame of feature for an experiment.
    """

    # this maps qaqc results to keys in the self.method_map. This is used for qaqc filtering
    qaqc_result_to_key = {
        "pca": "pca",
        "log_pca": "log_pca",
        "tsne": "tsne",
        "pearson_correlation": "pearson",
        "kendall_correlation": "kendall",
        "spearman_correlation": "spearman",
        "pearson_logtransformed_correlation": "log_pearson",
        "kendall_logtransformed_correlation": "log_kendall",
        "spearman_logtransformed_correlation": "log_spearman",
        "missing_feature_z_scores": "missing_feature_z_scores",
        "sum_intensity": "intensity_analysis",
        "mean_intensity": "intensity_analysis",
        "median_intensity": "intensity_analysis",
        "missing_dropped_sum_intensity": "intensity_analysis",
        "missing_dropped_mean_intensity": "intensity_analysis",
        "missing_dropped_median_intensity": "intensity_analysis",
        "log_missing_dropped_sum_intensity": "intensity_analysis",
        "log_missing_dropped_mean_intensity": "intensity_analysis",
        "log_missing_dropped_median_intensity": "intensity_analysis",
        "tics": "intensity_analysis",
        "log_tics": "intensity_analysis",
        "feature_count_z_scores": "feature_outlier_detection",
        "intensity_distribution": "intensity_distribution",
        "intensity_distribution_log": "intensity_distribution",
        "snr_distribution": "properties_distribution",
        "cSelectivity_distribution": "properties_distribution",
    }

    def __init__(self, feature_table, experiment, moniker):
        """
        This object wraps a feature table

        Args:
            feature_table_filepath (str): path to the feature table on disk
            experiment (Experiment object): the experiment object for this feature table
        """
        self.experiment = experiment
        self.feature_table = feature_table
        self.moniker = moniker

        #self.clean_columns()
        self.__mz_trees = {}
        self.__rt_trees = {}

        self.method_map = {
            "pca": self.pca,
            "tsne": self.tsne,
            "log_pca": partial(self.pca, log_transform=True),
            "pearson": partial(self.correlation_heatmap, correlation_type="pearson"),
            "kendall": partial(self.correlation_heatmap, correlation_type="kendall"),
            "spearman": partial(self.correlation_heatmap, correlation_type="spearman"),
            "log_pearson": partial(self.correlation_heatmap, correlation_type="pearson", log_transform=True),
            "log_kendall": partial(self.correlation_heatmap, correlation_type="kendall", log_transform=True),
            "log_spearman": partial(self.correlation_heatmap, correlation_type="spearman", log_transform=True),
            "missing_feature_percentiles": self.missing_feature_percentiles,
            "missing_feature_distribution": self.missing_feature_distribution,
            "missing_feature_z_scores": self.MissingFeatureZScores,
            "intensity_analysis": self.intensity_analysis,
            "feature_distribution": self.feature_distribution,
            "feature_outlier_detection": self.feature_distribution_outlier_detection,
            "intensity_distribution": self.intensity_distribution,
            "properties_distribution": self.properties_distribution,
        }

        self.qaqc_result_to_method = {
            k: self.method_map[v] for k, v in self.qaqc_result_to_key.items()
        }
        self.figure_params = None

    def get_mz_tree(self, mz_tol):
        """
        Construct an interval tree to search for features using a query
        mz and a specific mz tolerance in ppm.

        Args:
            mz_tol (float or int): float or int, this is the mass resolution in ppm

        Returns:
            intervaltree: interval tree for given mz_tol
        """
        if mz_tol not in self.__mz_trees:
            self.__mz_trees[mz_tol] = intervaltree.IntervalTree()
            for f_id, mz in self.feature_table[['id_number', 'mz']].values.tolist():
                mz_err = mz / 1e6 * mz_tol
                self.__mz_trees[mz_tol].addi(mz - mz_err, mz + mz_err, f_id)
        return self.__mz_trees[mz_tol]

    def get_rt_tree(self, rt_tol):
        """
        Construct an interval tree to search for features using a query
        rtime and a specific rtime tolerance in absolute units (sec).

        Args:
            rt_tol (float or int): this is the rtime tolerance in sec

        Returns:
            intervaltree: interval tree for given rt_tol
        """
        if rt_tol not in self.__rt_trees:
            self.__rt_trees[rt_tol] = intervaltree.IntervalTree()
            for f_id, rtime in self.feature_table[['id_number', 'rtime']].values.tolist():
                self.__rt_trees[rt_tol].addi(rtime - rt_tol, rtime + rt_tol, f_id)
        return self.__rt_trees[rt_tol]

    @property
    def sample_columns(self):
        """sample_columns

        Return a list of the column names in the feature table that are sample names.

        This is used when filtering the feature tables. When we search the experiment 
        for a set of samples with a given filter, this returns samples in the experiment
        that may not be in the feature table. We can use this list tofilter out the 
        samples in the experiment not in the feature table.

        Returns:
            list: list of sample columns

        """
        return self.feature_table.columns[11:]

    @property
    def non_sample_columns(self):
        """non_sample_columns

        Return a list of the column names in the feature table that are sample names.

        This is used when filtering the feature tables but typically the list of sample
        columns is used instead.

        Returns:
            list: list of columns that are not samples
        """
        return [x for x in self.feature_table.columns if x not in self.sample_columns]

    @property
    def log_transformed(self):
        """log_transformed

        This property queries the experiment object to determine if the feature table
        has been log transformed already

        Some operations log transform the feature table before analysis. Multiple log
        transforms would yield unwanted results so if an operation is going to log 
        transform a feature table, check this first to ensure that it is has not 
        already been log transformed.

        Returns:
            bool: true if table is log_transformed
        """
        return self.moniker in self.experiment.log_transformed_feature_tables

    @property
    def num_features(self):
        """
        Returns the number of features in the feature table

        Returns:
            int: number of features in feature table
        """
        return self.feature_table.shape[0] - 1

    @property
    def num_samples(self):
        """
        Returns the number of samples in the feature table

        Returns:
            int: number of samples in feature table
        """
        return len(self.sample_columns)

    @staticmethod
    def load(moniker, experiment):
        """
        This method yields a FeatureTable object when given a feature table moniker.
        FeatureTables are registered with the experiment object using a moniker, a 
        string that points to the file path for that feature table. This method 
        queries the experiment object, gets the feature table path, and creates
        the object.

        :param moniker: the string with which the FeatureTable is registered
        :type moniker: str
        :param experiment: the experiment object with the FeatureTable
        :type experiment: object
        :return: the feature table for the moniker
        :rtype: FeatureTable
        """

        df = pd.read_csv(experiment.feature_tables[moniker], sep="\t")
        mzml_to_name = {}
        for acquisition in experiment.acquisitions:
            if os.path.basename(acquisition.mzml_filepath).rstrip('.mzML') in df.columns:
                mzml_to_name[os.path.basename(acquisition.mzml_filepath).rstrip('.mzML')] = acquisition.name
        renamed = df.rename(columns=mzml_to_name)
        return FeatureTable(
            renamed,
            experiment,
            moniker,
        )

    def make_nonnegative(self, fill_value=1):
        """
        This replaces all NaN and 0 values in the feature table with the specified fill_value

        This is used primarially before log transforming the feature table to remove values 
        that cannot be log transformed

        :param fill_value: the value to replace NaN and 0 with, defaults to 1
        :type fill_value: int, optional
        """
        self.feature_table.fillna(0)
        for column in self.sample_columns:
            self.feature_table[column] = [max(x, fill_value) for x in self.feature_table[column]]

    def save(self, new_moniker=None, drop_invariants=True):
        """
        Save the feature table as a pandas-created .tsv and register the new on-disk location
        with the experiment object using the specified new_moniker or reuse the existing moniker. 
        By default this drops features that have no variance in the feature table. This can occur
        when a sample or samples are dropped and one or more features are zero or interpolated 
        only in the remaining samples.

        When an operation is performed that modifies a feature table, the resulting feature table 
        can be saved to disk using this method. The moniker for the feature table can be reused or 
        a new moniker provided. If a new moniker is provided it cannot be preferred or full since 
        we do not want to overwrite the asari results.

        Dropping invariants is recommended to reduce the size of the feature table and prevent 
        uninformative features from reaching downstream steps. There is no good reason to turn 
        it off, but the option exists.

        :param new_moniker: a new moniker to register the saved table with the experiment object, defaults to None
        :type new_moniker: string, optional
        :param drop_invariants: if true, drop features that have no variance, defaults to True
        :type drop_invariants: bool, optional
        """
        if new_moniker is None:
            new_moniker = self.moniker
            if new_moniker in {"preferred", "full"}:
                print("Cannot overwrite asari feature tables")
                sys.exit()

        if drop_invariants:
            self.drop_invariants()
        output_path = os.path.join(
            self.experiment.filtered_feature_tables_subdirectory,
            new_moniker + "_Feature_table.tsv",
        )
        self.feature_table.to_csv(
            os.path.join(
                self.experiment.filtered_feature_tables_subdirectory, output_path
            ),
            sep="\t",
            index=False,
        )
        self.experiment.feature_tables[new_moniker] = output_path
        self.experiment.save()
        if os.path.exists(self.experiment.qaqc_figs + "/" + new_moniker):
            utils.file_operations["delete"](self.experiment.qaqc_figs + "/" + new_moniker)

    def save_fig_path(self, name):
        """
        Given a desired name for a figure, this returns the path to which this figure should be
        saved.

        This ensures that the resulting path for the figure is a reasonable path without special 
        figures and is saved to the appropriate location in the experiment directory.

        :param name: desired name for the figure
        :type name: str
        :return: path to save figure
        :rtype: str
        """
        fig_path = os.path.join(
            os.path.abspath(self.experiment.qaqc_figs), self.moniker + "/"
        )
        if not os.path.exists(fig_path):
            os.makedirs(fig_path)

        name += json.dumps(self.figure_params['color_by']) + json.dumps(self.figure_params['marker_by']) + json.dumps(self.figure_params['text_by'])
        name = "".join(c for c in name if c.isalpha() or c.isdigit() or c==' ' or c=='_').rstrip()

        return os.path.join(fig_path, "_" + name + ".png")

    def gen_figure(
        self,
        figure_type,
        data,
        title="",
        x_label=None,
        y_label=None,
        fig_params=None,
        skip_annot=False,
        bins=100,
    ):
        """
        A single method is used to generate the figures for the FeatureTable. This allows for 
        consistent looking figures to be generated.

        The permitted types of figures are:

        "bar" - make a bar plot
        "scatter" - make a scatter plot
        "clustermap" - make a clustermap using seaborn
        "heatmap" - make a heatmap

        This will be refactored in the future but this method is responsible for generating
        all figures related to FeatureTables. The figure paramaters such as color, markers, 
        etc are stored as a datamember in the FeatureTable object. 

        :param figure_type: which figure type to make
        :type figure_type: str
        :param data: the data to plot
        :type data: can be dict or list (need to better document)
        :param title: the title for the figure, defaults to ''
        :type title: str, optional
        :param x_label: string to apply to the x-axis, defaults to None
        :type x_label: str, optional
        :param y_label: string to apply to the y-axis, defaults to None
        :type y_label: str, optional
        :param fig_params: if provided override the object's fig_param, defaults to None
        :type fig_params: dict, optional
        :param skip_annot: if true do not apply cosmetics to the figure, defaults to False
        :type skip_annot: bool, optional
        """

        if fig_params is None:
            fig_params = self.figure_params

        if fig_params["interactive"] or fig_params["save_figs"]:
            colors = fig_params["colors"]
            markers = fig_params["markers"]
            text = fig_params["text"]
            if figure_type == "scatter":
                if isinstance(data, dict):
                    xs = data.keys()
                    ys = data.values()
                else:
                    xs = data[:, 0]
                    ys = data[:, 1]
                plt.title(title)
                plt.xlabel(x_label)
                plt.ylabel(y_label)
                if skip_annot is False:
                    if markers and colors:
                        for x, y, c, m in zip(xs, ys, list(colors[0]), list(markers[0])):
                            plt.scatter(x, y, c=c, marker=m)
                    elif markers and not colors:
                        for x, y, m in zip(xs, ys, list(markers[0])):
                            plt.scatter(x, y, marker=m)
                    elif colors and not markers:
                        for x, y, c in zip(xs, ys, list(colors[0])):
                            plt.scatter(x, y, c=c)
                    else:
                        plt.scatter(xs, ys)
                    if text:
                        for x, y, t in zip(xs, ys, text[0]):
                            plt.text(x, y, t)
                else:
                    plt.scatter(xs, ys)
                if fig_params["marker_legend"] and skip_annot is False:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    handles = [
                        mlines.Line2D(
                            [],
                            [],
                            color="k",
                            marker=v,
                            linestyle="None",
                            markersize=10,
                            label=k,
                        )
                        for k, v in fig_params["marker_legend"].items()
                        if v in markers[0]
                    ]
                    handles += [
                        Patch(facecolor=v, label=k)
                        for k, v in fig_params["color_legend"].items()
                        if v in colors[0]
                    ]
                    plt.legend(
                        handles=handles,
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc="lower right",
                    )
            elif figure_type == "heatmap":
                if colors:
                    sns.clustermap(data, col_colors=colors, yticklabels=y_label)
                else:
                    sns.clustermap(data, yticklabels=y_label)
                plt.suptitle(title)
                if fig_params["color_legend"]:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    plt.legend(
                        [
                            Patch(facecolor=color)
                            for color in fig_params["color_legend"].values()
                        ],
                        list(fig_params["color_legend"].keys()),
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc="lower right",
                    )
            elif figure_type == "clustermap":
                if colors:
                    sns.clustermap(data, col_colors=colors)
                else:
                    sns.clustermap(data)
                plt.xticks
                plt.suptitle(title)
                if fig_params["color_legend"]:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    plt.legend(
                        [
                            Patch(facecolor=color)
                            for color in fig_params["color_legend"].values()
                        ],
                        list(fig_params["color_legend"].keys()),
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc="lower right",
                    )
            elif figure_type == "bar":
                if isinstance(data, dict):
                    data = [list(data.keys()), list(data.values())]
                if text and colors:
                    plt.bar(
                        [x + "_" + str(i) for i, x in enumerate(text[0])],
                        data[1],
                        color=colors[0],
                    )
                elif text and not colors:
                    plt.bar([x + "_" + str(i) for i, x in enumerate(text[0])], data[1])
                elif not text and colors:
                    plt.bar(list(range(len(data[1]))), data[1], color=colors[0])
                else:
                    plt.bar(list(range(len(data[1]))), data[1])
                plt.title(title)
                plt.xticks(rotation=90)
                plt.xlabel(title)
                plt.ylabel(x_label)
                if fig_params["color_legend"]:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    plt.legend(
                        [
                            Patch(facecolor=color)
                            for color in fig_params["color_legend"].values()
                        ],
                        list(fig_params["color_legend"].keys()),
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc="lower right",
                    )
            elif figure_type == "histogram":
                plt.suptitle(title)
                plt.hist(data, bins=bins)
            if fig_params["save_figs"]:
                


                plt.savefig(self.save_fig_path(title.replace(" ", "_")))
            if fig_params["interactive"]:
                plt.show()
            plt.clf()

    def search_for_feature(
        self, query_mz=None, query_rt=None, mz_tolerance=None, rt_tolerance=None
    ):
        """
        Given a query_mz and query_rt with corresponding tolerances in ppm and absolute units 
        respectively find all features by id_number that have a matching mz and rtime.

        All search fields are optional but if none are provided then all the features will be 
        considered matching. The mz tolerance should be in ppm while the rtime tolerance should 
        be provided in rtime units.

        :param query_mz: the mz to search for, defaults to None
        :type query_mz: float, optional
        :param query_rt: the rtime to search for, defaults to None
        :type query_rt: float, optional
        :param mz_tolerance: the tolerance in ppm for the mz match, defaults to None
        :type mz_tolerance: float, optional
        :param rt_tolerance: the tolerance in absolute units for the rt match, defaults to None
        :type rt_tolerance: float, optional
        :return: list of matching feature IDs
        :rtype: list
        """
        mz_matches, rt_matches = set(), set()
        if query_mz and mz_tolerance:
            mz_matches = {x.data for x in self.get_mz_tree(mz_tolerance).at(query_mz)}
            if query_rt is None or rt_tolerance is None:
                return mz_matches
        if query_rt and rt_tolerance:
            rt_matches = {x.data for x in self.get_rt_tree(rt_tolerance).at(query_rt)}
            if mz_matches is None or mz_tolerance is None:
                return rt_matches
        return list(rt_matches.intersection(mz_matches))

    def intensity_distribution(self, skip_zero=True):
        """
        This method generates various summaries of the intensity distribution in the feature table
        this includes TICs, LogTICs, median and mean intensity values including and excluding zeros 
        and including the values after log transforming the intensities.

        Args:
            skip_zero (bool, optional): if true, don't include zero values. Defaults to True.
        """
        if self.log_transformed:
            self.gen_figure(
                "histogram",
                [
                    x
                    for x in self.feature_table[self.sample_columns].values.flatten()
                    if x and skip_zero
                ],
                title="intensity_distribution_log",
                x_label="Intensity (Log-Transformed)",
                y_label="Counts",
            )
        else:
            self.gen_figure(
                "histogram",
                [
                    x
                    for x in self.feature_table[self.sample_columns].values.flatten()
                    if x and skip_zero
                ],
                title="intensity_distribution",
                x_label="Intensity",
                y_label="Counts",
            )
            self.gen_figure(
                "histogram",
                np.log2(
                    [
                        x
                        for x in self.feature_table[
                            self.sample_columns
                        ].values.flatten()
                        if x and skip_zero
                    ]
                ),
                title="intensity_distribution_log",
                x_label="Intensity (Log-Transformed)",
                y_label="Counts",
            )

    def properties_distribution(self):
        """
        This method generates figures for the distribution (a histogram) of every parameter in
        the feature table that is not id_number, parent_masstrack_id or actual intensities in 
        the samples. Useful for examining a feature table. 

        """
        for column in self.non_sample_columns:
            if column not in ["id_number", "parent_masstrack_id"]:
                self.gen_figure(
                    "histogram",
                    self.feature_table[column].values.flatten(),
                    title=column + "_distribution",
                    x_label=column,
                    y_label="Counts",
                    bins=100,
                )
                try:
                    self.gen_figure(
                        "histogram",
                        np.log10(
                            [
                                x
                                for x in self.feature_table[column].values.flatten()
                                if x > 0
                            ]
                        ),
                        title="log10_" + column + "_distribution",
                        x_label=column,
                        y_label="Counts",
                        bins=100,
                    )
                except TypeError:
                    pass
                except RuntimeWarning:
                    pass

    def median_correlation_outlier_detection(self, correlation_type="pearson"):
        """
        The median correlation of a sample against all other samples can be expressed as a z-score 
        against the median of ALL correlations in the experiment. A high or low Z-score indicates 
        that the sample was poorly correlated with other smaples in the experiment.

        Args:
            self: a feature table object
            correlation_type (str): can be 'pearson', 'spearman', 'kendall'

        Returns:
            dict: QAQC_result dict
        """
        correlation_result = self.correlation_heatmap(correlation_type=correlation_type, full_results=True)
        all_correlations = []
        median_correlations = {}
        for sample_name_1, corr_dict in correlation_result["Result"].items():
            correlation_for_sample_name_1 = []
            for sample_name_2, corr_value in corr_dict.items():
                if sample_name_1 != sample_name_2:
                    correlation_for_sample_name_1.append(corr_value)
            median_correlations[sample_name_1] = np.median(
                correlation_for_sample_name_1
            )
            all_correlations.extend(correlation_for_sample_name_1)
        all_correlations_std = np.std(all_correlations)
        all_correlations_median = np.median(all_correlations)
        z_score_correlations = {
            name: (median_correlation - all_correlations_median) / all_correlations_std
            for name, median_correlation in median_correlations.items()
        }

        self.gen_figure(
            "scatter",
            dict(enumerate(median_correlations.values())),
            title="Median Correlation Values for Samples",
            x_label="Sample",
            y_label="Median Correlation Value",
        )
        self.gen_figure(
            "scatter",
            dict(enumerate(z_score_correlations.values())),
            title="Median Correlation Z-Scores for Samples",
            x_label="Sample",
            y_label="Median Correlation Z-Score",
        )

        result = {
            "Type": "MedianCorrelationZScores",
            "Config": {},
            "Result": z_score_correlations,
        }
        return result

    def intensity_analysis(self):
        """
        This will report the sum, mean, median of features as well as
        those values when the missing values are removed or when they
        are log2 transformed.

        Returns:
            dict: QAQC_result dict
        """
        selected_ftable = self.feature_table[self.sample_columns].copy()
        intensity_sums = np.sum(selected_ftable, axis=0)
        mean_feature_intensity = np.mean(selected_ftable, axis=0)
        median_feature_intensity = np.median(selected_ftable, axis=0)

        selected_ftable = selected_ftable.copy()
        selected_ftable[selected_ftable == 0] = np.nan
        filtered_mean_feature_intensity = np.nanmean(selected_ftable, axis=0)
        filtered_median_feature_intensity = np.nanmedian(selected_ftable, axis=0)

        log_selected_ftable = np.log2(selected_ftable)
        log_filtered_intensity_sum = np.nansum(log_selected_ftable, axis=0)
        log_filtered_mean_feature_intensity = np.nanmean(log_selected_ftable, axis=0)
        log_filtered_median_feature_intensity = np.nanmedian(
            log_selected_ftable, axis=0
        )

        tics = np.nansum(selected_ftable, axis=0)
        log_tics = np.log2(tics)

        tables = [
            intensity_sums,
            mean_feature_intensity,
            median_feature_intensity,
            intensity_sums,
            filtered_mean_feature_intensity,
            filtered_median_feature_intensity,
            log_filtered_intensity_sum,
            log_filtered_mean_feature_intensity,
            log_filtered_median_feature_intensity,
            log_tics,
            tics,
        ]

        titles = [
            "sum_intensity",
            "mean_intensity",
            "median_intensity",
            "missing_dropped_sum_intensity",
            "missing_dropped_mean_intensity",
            "missing_dropped_median_intensity",
            "log_missing_dropped_sum_intensity",
            "log_missing_dropped_mean_intensity",
            "log_missing_dropped_median_intensity",
            "log_tics",
            "tics",
        ]

        for table, title in zip(tables, titles):
            results = dict(zip(self.sample_columns, table))
            self.gen_figure(
                "bar",
                results,
                title,
                x_label="title",
                y_label="sample",
            )

        result_values = {
            "sum_intensity": dict(zip(self.sample_columns, intensity_sums)),
            "mean_intensity": dict(zip(self.sample_columns, mean_feature_intensity)),
            "median_intensity": dict(zip(self.sample_columns, median_feature_intensity)),
            "missing_dropped_sum_intensity": dict(zip(self.sample_columns, intensity_sums)),
            "missing_dropped_mean_intensity": dict(zip(self.sample_columns, filtered_mean_feature_intensity)),
            "missing_dropped_median_intensity": dict(zip(self.sample_columns, filtered_median_feature_intensity)),
            "log_missing_dropped_sum_intensity": dict(zip(self.sample_columns, log_filtered_intensity_sum)),
            "log_missing_dropped_mean_intensity": dict(zip(self.sample_columns, log_filtered_mean_feature_intensity)),
            "log_missing_dropped_median_intensity": dict(zip(self.sample_columns, log_filtered_median_feature_intensity)),
            "log_tics": dict(zip(self.sample_columns, log_tics)),
            "tics": dict(zip(self.sample_columns, tics)),
        }
        results = []
        for k, v in result_values.items():
            results.append({"Type": k, "Config": {}, "Result": v})
        return results

    def correlation_heatmap(self, correlation_type, log_transform=False, full_results=False):
        """correlation_heatmap

        Using a specified correlation function generate a correlation heatmap for the feature 
        table. Optionally, log transform the feature table first.

        The permitted correlation types are:

        "pearson", "spearman" or "kendall"

        Only pearson will log_transform the feature table if enabled since the non-parametric 
        correlations will not be affected by the log transform.

        :param figure_params: dictionary with the figure params
        :type figure_params: dict
        :param correlation_type: what correlation type to use
        :type correlation_type: str
        :param log_transform: if true, log transform before linear correlation, defaults to True
        :type log_transform: bool, optional
        :param full_results: if true, yield the corr matrix as dictionary, else discard the matrix
        :return: a dict with the correlation results and configuration used to generate result
        :rtype: dict
        """
        corr_method = utils.correlation_modes[correlation_type]
        corr_matrix = np.zeros((self.num_samples, self.num_samples))
        working_table = self.feature_table.copy()
        if log_transform:
            working_table = np.log2(working_table[self.sample_columns] + 1)
        for i, s1 in enumerate(self.sample_columns):
            val_s1 = working_table[s1]
            for j, s2 in enumerate(self.sample_columns):
                if corr_matrix[j][i] != 0:
                    corr_matrix[i][j] = corr_matrix[j][i]
                else:
                    corr = corr_method(val_s1, working_table[s2])
                    if hasattr(corr, 'statistic'):
                        corr_matrix[i][j] = corr.statistic
                    else:
                        corr_matrix[i][j] = corr[0][1]
        if log_transform:
            title = correlation_type + "_logtransformed_correlation"
        else:
            title = correlation_type + "_correlation"

        self.gen_figure(
            "clustermap",
            corr_matrix,
            title=title,
            x_label=self.figure_params["text"],
            y_label=self.figure_params["text"],
        )
        if full_results:
            result = {
                "Type": title,
                "Config": {"Metric": correlation_type, "LogTransformed": log_transform},
                "Result": {
                    self.sample_columns[i]: {
                        self.sample_columns[j]: float(corr_matrix[i][j])
                        for j in range(corr_matrix.shape[0])
                    }
                    for i in range(corr_matrix.shape[0])
                },
            }
        else:
            result = {
                "Type": title,
                "Config": {"Metric": correlation_type, "LogTransformed": log_transform},
                "Result": {"CorrMatrix": corr_matrix, "Samples": self.sample_columns}
            }
        return result

    def pca(self, log_transform=False):
        """
        Perform PCA on provided feature table, optionally log transform
        it first.

        Args:
            log_transform (bool, optional): if true log2 transform the table

        Returns:
            dict: QAQC_result dict
        """
        sample_ftable = self.feature_table[self.sample_columns].T.copy()
        scaler = StandardScaler()
        pca_embedder = PCA(n_components=2)
        pca_log_transformed = False
        if log_transform and not self.log_transformed:
            pca_log_transformed = True
            sample_ftable = np.log2(sample_ftable + 1)
        pca_embedding = pca_embedder.fit_transform(
            scaler.fit_transform((sample_ftable))
        )
        title = "pca" if not pca_log_transformed else "log_pca"
        self.gen_figure(
            "scatter",
            pca_embedding,
            title,
            x_label="PC 1 "
            + str(round(pca_embedder.explained_variance_ratio_[0] * 100, 1))
            + "%",
            y_label="PC 2 "
            + str(round(pca_embedder.explained_variance_ratio_[1] * 100, 1))
            + "%",
        )
        result = {
            "Type": "pca" if not pca_log_transformed else "log_pca",
            "Config": {"n_components": 2, "scaler": "StandardScaler"},
            "Result": {
                "Sample_Coord_Dict": {
                    name: list(coord)
                    for name, coord in zip(self.sample_columns, pca_embedding)
                }
            },
        }
        return result

    def tsne(self, perplexity=30):
        """
        Perform TSNE on provided feature table

        Args:
            perplexity (int): perplexity value for TSNE

        Results
            dict: QAQC result dict
        """
        try:
            tnse_embedded_vector_matrix = TSNE(
                n_components=2, perplexity=perplexity
            ).fit_transform(self.feature_table[self.sample_columns].T)
            self.gen_figure(
                "scatter",
                tnse_embedded_vector_matrix,
                "tsne",
                x_label="Latent 1",
                y_label="Latent 2",
            )
            result = {
                "Type": "tsne",
                "Config": {"n_components": 2},
                "Result": {
                    "Sample_Coord_Dict": {
                        name: [float(x) for x in coord]
                        for name, coord in zip(
                            self.sample_columns, tnse_embedded_vector_matrix
                        )
                    }
                },
            }
            return result
        except:
            if perplexity > 0:
                perplexity -= 1
                return self.tsne(perplexity)
            return {}

    def missing_feature_percentiles(self):
        """
        Calculate the distribution of missing features with respect to percent of smaples with 
        feature

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            interactive_plot (bool, optional): if True, interactive plots are made. 
                Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """

        def __count_feature(row, columns):
            return np.sum([1 for x in row[columns] if x > 0])

        num_sample_with_feature = self.feature_table.apply(
            __count_feature, axis=1, args=(self.sample_columns,)
        )
        percentile_table = []
        for percentile in range(101):
            num_samples_threshold = len(self.sample_columns) * percentile / 100
            percentile_table.append(
                [
                    percentile,
                    num_samples_threshold,
                    int(np.sum(num_sample_with_feature <= num_samples_threshold)),
                ]
            )
        self.gen_figure(
            "scatter",
            np.array([[x[0], x[2]] for x in percentile_table]),
            title="Missing Feature Percentiles",
            x_label="Percentile",
            y_label="Num. Dropped Features",
            skip_annot=True,
        )
        result = {
            "Type": "missingfeaturepercentiles",
            "Config": {},
            "Result": {"PercentileTable": percentile_table},
        }
        return result

    def missing_feature_distribution(self, intensity_cutoff=0):
        """
        Count the number of missing features or featuers below the specified intensity cutoff per 
        features

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values below this intesnity are considered missing. 
                Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. 
                Defaults to False.

        Returns:
            dict: dictionary storing the result of this QCQA operation
        """
        masked_ftables = self.feature_table[self.sample_columns] <= intensity_cutoff
        missing_feature_counts = dict(
            zip(self.sample_columns, [0 for _ in self.sample_columns])
        )
        for name in self.sample_columns:
            for value in masked_ftables[name]:
                if value is True:
                    missing_feature_counts[name] += 1
        self.gen_figure(
            "bar",
            (
                self.sample_columns,
                [missing_feature_counts[name] for name in self.sample_columns],
            ),
            title="missing_feature_counts",
            x_label="Missing Feature Counts",
            y_label="Num. Missing Features",
        )
        result = {
            "Type": "MissingFeatureDistribution",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {
                name: int(num_missing)
                for name, num_missing in missing_feature_counts.items()
            },
        }
        return result

    def feature_distribution(self, intensity_cutoff=0):
        """
        Count the number of features above the specified intensity cutoff per features

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values with greater intensiy are considered. 
                Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. 
                Defaults to False.

        Returns:
            dict: dictionary storing the result of this QCQA operation
        """
        masked_ftables = self.feature_table[self.sample_columns] > intensity_cutoff
        feature_counts = dict(
            zip(self.sample_columns, [0 for _ in self.sample_columns])
        )
        for name in self.sample_columns:
            for value in masked_ftables[name]:
                if value is True:
                    feature_counts[name] += 1
        self.gen_figure(
            "bar",
            (
                self.sample_columns,
                [feature_counts[name] for name in self.sample_columns],
            ),
            title="Feature Counts",
            y_label="Num. Features",
        )
        result = {
            "Type": "FeatureDistribution",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {
                name: int(num_missing) for name, num_missing in feature_counts.items()
            },
        }
        return result

    def feature_distribution_outlier_detection(self, intensity_cutoff=0):
        """
        Count the number of features above the specified intensity cutoff per features and express
        as a Z-score based on feature count across all samples.

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values above this intensity are considered. 
                Defaults to 0.
            interactive_plot (bool, optional): if True, plots are interactive. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        feature_counts_result = self.feature_distribution(
            intensity_cutoff=intensity_cutoff
        )
        feature_counts = np.array([*feature_counts_result["Result"].values()])
        feature_z_scores = (feature_counts - np.mean(feature_counts)) / np.std(
            feature_counts
        )
        self.gen_figure(
            "scatter",
            dict(enumerate(feature_z_scores)),
            title="feature_count_z_scores",
            x_label="Sample",
            y_label="Num Feature Z-Score",
        )
        result = {
            "Type": "feature_count_z_scores",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {
                name: float(z_score)
                for name, z_score in zip(self.sample_columns, feature_z_scores)
            },
        }
        return result

    def MissingFeatureZScores(self, intensity_cutoff=0):
        """
        Count the number of features below the specified intensity cutoff per features and express
        as a Z-score based on missing feature count across all samples.

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values below this intensity are considered missing. 
                Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. 
                Defaults to False.

        Returns:
            dict: dictionary storing the result of this QCQA operation
        """
        missing_feature_counts_result = self.missing_feature_distribution(
            intensity_cutoff=intensity_cutoff
        )
        # this relies upon the sorted order of the dictionary, may not be safe in all Python versions
        sample_names = [*missing_feature_counts_result["Result"].keys()]
        missing_feature_counts = np.array([*missing_feature_counts_result["Result"].values()])
        missing_feature_z_scores = (missing_feature_counts - np.mean(missing_feature_counts)) / np.std(missing_feature_counts)
        self.gen_figure(
            "scatter",
            dict(enumerate(missing_feature_z_scores)),
            title="missing_feature_z_scores",
            x_label="Sample",
            y_label="Num Missing Feature Z-Score",
        )
        result = {
            "Type": "missing_feature_z_scores",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {
                name: float(z_score)
                for name, z_score in zip(sample_names, missing_feature_z_scores)
            },
        }
        return result

    def drop_invariants(self, zeros_only=False):
        """
        This method drops features that have all zero intensity or the same intensity across all 
        samples.

        This situation occurs as a result of filtering. For instance if a contaiminant is only 
        seen in the blanks, when the blanks are dropped from the feature table, that feature is
        still in the table but will be zero (or an interpolated value) for the remaning samples. 
        These features have no information and can complicate downstream analysis.

        :param zeros_only: if true, only drop features that are all zero, defaults to False
        :type zeros_only: bool, optional
        """

        def __filter_invariant(row, columns):
            values = set()
            for column in columns:
                if column in row:
                    values.add(row[column])
            values = list(values)
            if len(values) == 1:
                if zeros_only and values[0] == 0:
                    return False
                return False
            return True

        to_keep = []

        for keep_feature, id_number in zip(
            self.feature_table.apply(
                __filter_invariant, axis=1, args=(self.sample_columns,)
            ),
            self.feature_table["id_number"],
        ):
            if keep_feature:
                to_keep.append(id_number)
        self.feature_table = self.feature_table[
            self.feature_table["id_number"].isin(to_keep)
        ].copy()

        for sample_column in self.sample_columns:
            unique_values = set()
            for value in self.feature_table[sample_column]:
                unique_values.add(value)
            unique_values = list(unique_values)
            if len(unique_values) == 1:
                if zeros_only and unique_values[0] == 0:
                    self.feature_table.drop(columns=[sample_column], inplace=True)
                else:
                    self.feature_table.drop(columns=[sample_column], inplace=True)

    def drop_sample_by_name(self, drop_name, drop_others=False):
        """
        This method drops a sample from a feature table by its name.x

        Optionally all other samples that do not match the name can be dropped as well.

        Args:
            drop_name (_type_): the name to be dropped
            drop_others (bool, optional): drop other samples if true. Defaults to False.
        """
        starting_columns = list(self.feature_table.columns)
        if drop_others:
            self.feature_table.drop(
                columns=[x for x in self.sample_columns if x != drop_name], inplace=True
            )
        else:
            self.feature_table.drop(columns=drop_name, inplace=True)
        print("Dropping:")
        for x in starting_columns:
            if x not in self.feature_table.columns:
                print("\t", x)
            
            

    def drop_samples_by_filter(self, sample_filter, drop_others=False):
        """
        Given a sample filter, a dictionary as described elsewhere, drop all other samples.

        Args:
            sample_filter (dict): the dictionary specifying the filter 
            drop_others (bool, optional): if true, reverse the logic of the drop. Defaults to False.
        """
        to_drop = [acq.name for acq in self.experiment.filter_samples(sample_filter)]
        to_drop = [x for x in to_drop if x in self.sample_columns]
        do_not_drop = [x for x in self.sample_columns if x not in to_drop]
        if drop_others:
            to_drop, do_not_drop = do_not_drop, to_drop
        print("Dropping:")
        for x in to_drop:
            print("\t", x)
        self.feature_table.drop(columns=to_drop, inplace=True)

    def drop_samples_by_field(self, value, field, drop_others=False):
        """
        For a given field and a value for that field drop all samples that match or all samples
        that do not match. 

        Args:
            value (str): the value for the field to be dropped
            field (str): the field corresponding to the value that needs to be dropped
            drop_others (bool, optional): if true drop samples that do not match. Defaults to False.
        """
        self.drop_samples_by_filter(
            {field: {"includes": [value]}}, drop_others=drop_others
        )

    def drop_samples_by_qaqc(self, qaqc_filter, drop_others=False, params=None):
        """
        This drops samples based on a qaqc result. This requires an additional
        field in the filter called "conditions" which can accept keys ">" and
        "<" that control the logic of the comparison. Currently only numerical
        metrics can be used for dropping. The "Action" field is also need and can
        accept the values "Keep" and "Drop" which specify what should happen to 
        the sample that matches the filter. 

        The permitted qaqc results for this filter are described in self.qaqc_results_to_key
        and if the metric has not been evaluated, it will be evaluated on demand 
        in this method. 

        #todo - params seems unnecessary here

        Args:
            qaqc_filter (dict): a dict detailing the qaqc filter
            drop_others (bool, optional): if true, reverse the logic of the drop. Defaults to False.
            params (dict, optional): the params from main, needed for figure_params. 
                Defaults to None.
        """
        to_drop = []
        max_value, min_value = np.inf, -np.inf
        for field in qaqc_filter:
            if ">" in qaqc_filter[field]["Conditions"]:
                max_value = float(qaqc_filter[field]["Conditions"][">"])
            if "<" in qaqc_filter[field]["Conditions"]:
                min_value = float(qaqc_filter[field]["Conditions"]["<"])
            if self.moniker not in self.experiment.qcqa_results:
                self.experiment.qcqa_results[self.moniker] = {}
            if field not in self.experiment.qcqa_results[self.moniker] and params:
                method = self.qaqc_result_to_method.get(field, None)
                self.figure_params = {}
                self.figure_params["interactive"] = False
                self.figure_params["save_figs"] = False
                if method:
                    result = method()
                    if isinstance(result, dict):
                        result = [result]
                    for qaqc_result in result:
                        qcqa_results = self.experiment.qcqa_results[self.moniker]
                        qcqa_results[qaqc_result["Type"]] = qaqc_result
                        self.experiment.qcqa_results[self.moniker][qaqc_result["Type"]] = qaqc_result
                else:
                    print("No method found for " + field)
            qaqc_results_for_field = self.experiment.qcqa_results[self.moniker].get(field, None)
            if qaqc_results_for_field:
                for sample, value in qaqc_results_for_field["Result"].items():
                    if not min_value < float(value) < max_value:
                        if qaqc_filter[field]["Action"] == "Keep":
                            pass
                        elif qaqc_filter[field]["Action"] == "Drop":
                            to_drop.append(sample)
            else:
                print("No qaqc results found for " + field)
        to_drop = [x for x in to_drop if x in self.sample_columns]
        if drop_others:
            to_drop = [x for x in self.sample_columns if x not in to_drop]
        print("Dropping:")
        if to_drop:
            for x in to_drop:
                print(x)
            self.feature_table.drop(columns=to_drop, inplace=True)
        print("\n".join(to_drop))

    def blank_mask(
        self,
        blank_value="Blank",
        sample_value="Unknown",
        query_field="Sample Type",
        blank_intensity_ratio=3,
        by_batch=None,
        logic_mode="or",
    ):
        """blank_mask

        Given a feature table containing samples that we consider blanks, drop all features in 
        non-blank samples that do not have an intensity blank_intensity_ratio times higher than 
        the mean intensity in the blanks.

        The blank samples are specified by the comibnation of blank_type and type_field. Non-blank
        samples are specified by sample_type and type_field in a similar manner.

        If there are batches in the experiment, blank masking is done per-batch. Then dropped if 
        the ratio condition is not true in one sample (if logic_mode is "or") or in all samples if
        logic_mode is "and". The batches are specified given a field in the metadata via the by_batch field.

        _extended_summary_

        :param by_batch: if true, blank mask by the batch field, defaults to None
        :type by_batch: str, optional
        :param blank_intensity_ratio: sample feautre intensity / blank intensity must exceed this value to be kept, defaults to 3
        :type blank_intensity_ratio: int, optional
        :param logic_mode: determines if a feature is dropped if it fails the test in one batch or all batches, defaults to "or"
        :type logic_mode: str, optional
        :param blank_type: the value of type_field that specifies the blanks, defaults to "Blank"
        :type blank_type: str, optional
        :param sample_type: the value of type_field that specifies the study samples, defaults to "Unknown"
        :type sample_type: str, optional
        :param type_field: the field to look for the sample type in, defaults to "Sample Type"
        :type type_field: str, optional
        """

        def __non_zero_mean(row, columns):
            non_zero_columns = [x for x in row[columns] if x > 0]
            return np.mean(non_zero_columns) if len(non_zero_columns) > 0 else 0

        def __any_logical(row, columns):
            return np.any(row[columns] == True)

        def __all_logical(row, columns):
            return np.all(row[columns] == True)

        blanks = self.experiment.filter_samples({query_field: {"includes": [blank_value]}})
        samples = self.experiment.filter_samples({query_field: {"includes": [sample_value]}})
        blank_names = [x.name for x in blanks if x.name in self.sample_columns]
        sample_names = [x.name for x in samples if x.name in self.sample_columns]

        blank_mask_columns = []
        if by_batch:
            for batch_name, batch_name_list in self.experiment.batches(
                by_batch
            ).items():
                batch_blanks = [x for x in batch_name_list if x in blank_names]
                batch_samples = [x for x in batch_name_list if x in sample_names]
                blank_means = self.feature_table.apply(
                    __non_zero_mean, axis=1, args=(batch_blanks,)
                )
                sample_means = self.feature_table.apply(
                    __non_zero_mean, axis=1, args=(batch_samples,)
                )
                to_filter = []
                for blank_mean, sample_mean in zip(blank_means, sample_means):
                    to_filter.append(blank_mean * blank_intensity_ratio > sample_mean)
                blank_mask_column = "blank_masked_" + batch_name
                blank_mask_columns.append(blank_mask_column)
                self.feature_table[blank_mask_column] = to_filter
            if logic_mode == "and":
                self.feature_table["mask_feature"] = self.feature_table.apply(
                    __all_logical, axis=1, args=(blank_mask_columns,)
                )
            elif logic_mode == "or":
                self.feature_table["mask_feature"] = self.feature_table.apply(
                    __any_logical, axis=1, args=(blank_mask_columns,)
                )
            for blank_mask_column in blank_mask_columns:
                self.feature_table.drop(columns=blank_mask_column, inplace=True)
        else:
            blank_means = self.feature_table.apply(
                __non_zero_mean, axis=1, args=(list(blank_names),)
            )
            sample_means = self.feature_table.apply(
                __non_zero_mean, axis=1, args=(list(sample_names),)
            )
            to_filter = []
            for blank_mean, sample_mean in zip(blank_means, sample_means):
                to_filter.append(blank_mean * blank_intensity_ratio > sample_mean)
            blank_mask_column = "mask_feature"
            self.feature_table["mask_feature"] = to_filter
        self.feature_table = self.feature_table[
            self.feature_table["mask_feature"] == False
        ]
        self.feature_table.drop(columns="mask_feature", inplace=True)

    def impute_missing_features(self, ratio=0.5, by_batch=None, method="min"):
        """impute_missing_features 

        Fill zero values with a small value to make downstream stats more robust. This value is 
        a multiplier of the minimum value for that feature observed across all samples, excluding
        zeros.

        :param ratio: multiply min value by this value, defaults to 0.5
        :type ratio: float, optional
        :param by_batch: if try, impute per batch, defaults to None
        :type by_batch: str, optional
        """

        def __calc_impute_value(row, sample_names):
            values = [x for x in row[sample_names] if x > 0]
            if values:
                return utils.descriptive_stat_modes[method](values) * ratio
            return 0

        if by_batch:
            for _, b_sample_names in self.experiment.batches(by_batch).values():
                b_sample_names = list(set(b_sample_names).intersection(set(self.sample_columns)))
                i_v = self.feature_table.apply(__calc_impute_value, axis=1, args=(b_sample_names,))
                self.feature_table["interp_value"] = i_v
                for sample_name in b_sample_names:
                    interp_values = self.feature_table[[sample_name, "interp_value"]].max(axis=1)
                    self.feature_table[sample_name] = interp_values
                self.feature_table.drop(columns="interp_value", inplace=True)
        else:
            i_v = self.feature_table.apply(__calc_impute_value, axis=1, args=(self.sample_columns,))
            self.feature_table["interp_value"] =  i_v
            for sample_name in self.sample_columns:
                interp_values = self.feature_table[[sample_name, "interp_value"]].max(axis=1)
                self.feature_table[sample_name] = interp_values
            self.feature_table.drop(columns="interp_value", inplace=True)

    def TIC_normalize(
        self, tic_normalization_percentile=0.90, by_batch=None, normalize_mode="median"
    ):
        """TIC_normalize 

        This method will normalize the features of each acquisition based on the TICs of 
        the samples. In this case, the TICs are calculated only using features that are 
        present in TIC_normalization_percentile or greater percent of the samples. 

        Normalize mode determines how the normalization factor will be calculated, using
        either the mean or the median. 

        If by_batch is given, the normalization is performed in batches first with the 
        batches determined by the field specified by_batch. Then all batches are normalized 
        to one another. 

        :param TIC_normalization_percentile: only features in more than this 
        percent of samples are used for TIC calcualtion, defaults to 0.90
        :type TIC_normalization_percentile: float
        :param by_batch: the field on which to group samples into batches
        :type by_batch: str, optional

        :param normalize_mode: the method used to calculate the normalization factors,
          defaults to 'median'
        :type normalize_mode: str, optional
        """

        if by_batch is not None:
            aggregate_batch_tics = {}
            for batch_name, batch_name_list in self.experiment.batches(
                by_batch
            ).items():
                batch_name_list = [
                    x for x in batch_name_list if x in self.feature_table.columns
                ]
                self.feature_table["percent_inclusion"] = np.sum(
                    self.feature_table[batch_name_list] > 0, axis=1
                ) / len(batch_name_list)
                tics = {
                    sample: np.sum(
                        self.feature_table[
                            self.feature_table["percent_inclusion"]
                            > tic_normalization_percentile
                        ][sample]
                    )
                    for sample in batch_name_list
                }
                norm_factors = {
                    sample: utils.descriptive_stat_modes[normalize_mode](
                        list(tics.values())
                    )
                    / value
                    for sample, value in tics.items()
                }
                aggregate_batch_tics[batch_name] = utils.descriptive_stat_modes[
                    normalize_mode
                ](list(tics.values()))
                for sample, norm_factor in norm_factors.items():
                    self.feature_table[sample] = (
                        self.feature_table[sample] * norm_factor
                    )
            aggregate_batch_tic_corrections = {
                batch: utils.descriptive_stat_modes[normalize_mode](
                    list(aggregate_batch_tics.values())
                )
                / value
                for batch, value in aggregate_batch_tics.items()
            }
            for batch_name, batch_name_list in self.experiment.batches(
                by_batch
            ).items():
                batch_name_list = [
                    x for x in batch_name_list if x in self.feature_table.columns
                ]
                for sample in batch_name_list:
                    self.feature_table[sample] = (
                        self.feature_table[sample]
                        * aggregate_batch_tic_corrections[batch_name]
                    )
        else:
            sample_names = [
                x
                for x in self.feature_table.columns
                if x in [a.name for a in self.experiment.acquisitions]
            ]
            self.feature_table["percent_inclusion"] = np.sum(
                self.feature_table[sample_names] > 0, axis=1
            ) / len(sample_names)
            tics = {
                sample: np.sum(
                    self.feature_table[
                        self.feature_table["percent_inclusion"]
                        > tic_normalization_percentile
                    ][sample]
                )
                for sample in sample_names
            }
            norm_factors = {
                sample: utils.descriptive_stat_modes[normalize_mode](
                    list(tics.values())
                )
                / value
                for sample, value in tics.items()
            }
            for sample, norm_factor in norm_factors.items():
                self.feature_table[sample] = self.feature_table[sample] * norm_factor
        self.feature_table.drop(columns="percent_inclusion", inplace=True)

    def batch_correct(self, by_batch):
        """
        This method batch corrects the feature intensities. The 
        batches are determined dynamically using the by_batch field. 

        :param by_batch: the field on which to batch sampels
        :type by_batch: str
        """
        if len(self.experiment.batches(by_batch).keys()) > 1:
            batch_idx_map = {}
            for batch_idx, (_, acquisition_list) in enumerate(self.experiment.batches(by_batch).items()):
                for acquisition in acquisition_list:
                    batch_idx_map[acquisition] = batch_idx
            batches = [batch_idx_map[x] for x in self.sample_columns]
            batch_corrected = pycombat(self.feature_table[self.sample_columns], batches)
            for column in batch_corrected.columns:
                self.feature_table[column] = batch_corrected[column]
            self.make_nonnegative(fill_value=1)
        else:
            print("Unable to batch correct if only one batch!")
            sys.exit()

    def log_transform(self, log_mode="log2"):
        """
        log transform the features in the table.

        :param log_mode: can be log10 or log2, which type of log to use, defaults to "log2"
        :type log_mode: str, optional
        """
        for sample_name in self.sample_columns:
            self.feature_table[sample_name] = utils.log_modes[log_mode](self.feature_table[sample_name] + 1)

    def drop_missing_features(
        self, by_batch=None, drop_percentile=0.8, logic_mode="or"
    ):
        """drop_missing_features 
        
        This method will drop features that are uncommon in the feature table.

        Drop_percentile is the threshold for inclusion.

        :param by_batch: if provided, perform the operation on each batch separately. with 
        batches defined by this field., defaults to None
        :type by_batch: str, optional
        :param drop_percentile: features present in this percent or fewer of samples are dropped
        , defaults to 0.8
        :type drop_percentile: float, optional
        :param logic_mode: if by batch, drop any feature that fails the threshold in 'any' batch 
        or 'all' batches, defaults to "or"
        :type logic_mode: str, optional
        """

        def __any(row, columns, drop_percentile):
            return not np.any(row[columns] >= drop_percentile)

        def __all(row, columns, drop_percentile):
            return not np.all(row[columns] >= drop_percentile)

        batch_columns = []
        if by_batch:
            for batch_name, batch_name_list in self.experiment.batches(
                by_batch
            ).items():
                batch_column = "percent_inclusion_" + batch_name
                filtered_batch_name_list = [
                    x for x in batch_name_list if x in self.sample_columns
                ]
                self.feature_table[batch_column] = np.sum(
                    self.feature_table[filtered_batch_name_list] > 0, axis=1
                ) / len(filtered_batch_name_list)
                batch_columns.append(batch_column)
            if logic_mode == "and":
                self.feature_table["drop_feature"] = self.feature_table.apply(
                    __all, axis=1, args=(batch_columns, drop_percentile)
                )
            elif logic_mode == "or":
                self.feature_table["drop_feature"] = self.feature_table.apply(
                    __any, axis=1, args=(batch_columns, drop_percentile)
                )
        else:
            self.feature_table["drop_feature"] = (
                np.sum(self.feature_table[self.sample_columns] > 0, axis=1)
                / len(self.sample_columns)
            ) < drop_percentile
        self.feature_table = self.feature_table[
            self.feature_table["drop_feature"] == False
        ]
        self.feature_table.drop(columns="drop_feature", inplace=True)

    def __gen_color_cosmetic_map(self, colorby, seed=None):
        """
        This method generates the cosmetic map for the fields in colorby. Essentially, 
        this is a mapping of values for the fiels in colorby to colors for plotting.

        Args:
            colorby (list): list of fields that need colors
            seed (int, optional): if provided, this sets the seed for RNG purposes. Should allow reproducible maps. 
            Defaults to None.

        Returns:
            dict: map of values to colors
        """
        color_cosmetic_map = {}
        for color in colorby:
            cosmetic_map = self.experiment.generate_cosmetic_map(color, "colors", seed)
            color_cosmetic_map.update({("colors", k): v for k, v in cosmetic_map.items()})
        return color_cosmetic_map

    def __gen_marker_cosmetic_map(self, markerby, seed=None):
        """
        This method generates the cosmetic map for the fields in markerby. Essentially, 
        this is a mapping of values for the fiels in markerby to markers for plotting.

        Args:
            colorby (list): list of fields that need markers
            seed (int, optional): if provided, this sets the seed for RNG purposes. Should allow reproducible maps. 
            Defaults to None.

        Returns:
            dict: map of values to markers
        """
        marker_cosmetic_map = {}
        for marker in markerby:
            cosmetic_map = self.experiment.generate_cosmetic_map(marker, "markers", seed)
            marker_cosmetic_map.update({("markers", k): v for k, v in cosmetic_map.items()})
        return marker_cosmetic_map

    def generate_cosmetic(self, colorby=None, markerby=None, textby=None, seed=None):
        """generate_cosmetic

        Plots need colors, markers, and text fields. The colors and markers need to defined
        on the fly since they may not be known a priori. This method generates this mapping
        based on the fields in coloryb, markerby and textby. 

        :param colorby: list of fields that need colors, defaults to None
        :type colorby: list, optional
        :param markerby: list of fields that need markers, defaults to None
        :type markerby: list, optional
        :param textby: list of fields to be used for text, defaults to None.
        largely here for future expansion
        :type textby: list, optional
        :param seed: if provided, this sets the seed for RNG purposes. Should allow reproducible maps. 
            Defaults to None.
        :type seed: int, optional
        :return: map of field values to colors, markers and text
        :rtype: dict
        """
        combined_cosmetic_map = {}
        combined_cosmetic_map.update(self.__gen_color_cosmetic_map(colorby, seed))
        combined_cosmetic_map.update(self.__gen_marker_cosmetic_map(markerby, seed))
        cosmetics = {
            "colors": [[] for _ in colorby],
            "markers": [[] for _ in markerby],
            "texts": [[] for _ in textby]
        }
        legends = {
            "colors": {},
            "markers": {}
        }
        acq_name_map = {acq.name: acq for acq in self.experiment.acquisitions}
        input_name_map = {os.path.basename(acq.input_file): acq for acq in self.experiment.acquisitions}
        for sample_name in self.sample_columns:
            if sample_name in acq_name_map:
                acquisition = acq_name_map[sample_name]
            elif sample_name + ".mzML" in acq_name_map:
                acquisition = acq_name_map[sample_name + ".mzML"]
            elif sample_name in input_name_map:
                acquisition = input_name_map[sample_name]
                
            

            #acquisition = acq_name_map[sample_name.split("___")[-1]]
            for i, x in enumerate(colorby):
                value_for_cosmetic = acquisition.metadata_tags[x]
                cosmetic_for_value = combined_cosmetic_map[("colors", value_for_cosmetic)]
                cosmetics["colors"][i].append(cosmetic_for_value)
                legends["colors"][value_for_cosmetic] = cosmetic_for_value
            for i, x in enumerate(markerby):
                value_for_cosmetic = acquisition.metadata_tags[x]
                cosmetic_for_value = combined_cosmetic_map[("markers", value_for_cosmetic)]
                cosmetics["markers"][i].append(cosmetic_for_value)
                legends["markers"][value_for_cosmetic] = cosmetic_for_value
            for i, x in enumerate(textby):
                cosmetics["texts"][i].append(acquisition.metadata_tags[x])
        cos_colors, cos_markers, cos_texts = [cosmetics[x] for x in ["colors", "markers", "texts"]]
        leg_colors, leg_markers = [legends[x] for x in ["colors", "markers"]]
        return cos_colors, cos_markers, cos_texts, leg_colors, leg_markers

    def generate_figure_params(self, params):
        """
        This method generates the parameters used for plotting. 

        Args:
            params (dict): the params passed on the CLI. 
        """
        for x in ["color_by", "marker_by", "text_by"]:
            if x in params and isinstance(params[x], str):
                params[x] = json.loads(params[x])
        colors, markers, texts, color_legend, marker_legend = self.generate_cosmetic(
            params["color_by"], params["marker_by"], params["text_by"], params["seed"]
        )
        self.figure_params = {
            "acquisitions": list(self.sample_columns),
            "interactive": params["interactive_plots"],
            "save_figs": params["save_plots"],
            "text": texts,
            "markers": markers,
            "colors": colors,
            "color_legend": color_legend,
            "marker_legend": marker_legend,
            "color_by": params["color_by"],
            "marker_by": params["marker_by"],
            "text_by": params["text_by"]
        }

    def QAQC(self, params):
        """
        This is the wrapper for all the qcqa functions.

        If these fields are present in the params, it will determine which methods are performed:

            pca (bool, optional): Defaults to False.
            tsne (bool, optional): Defaults to False.
            pearson (bool, optional): Defaults to False.
            spearman (bool, optional): Defaults to False.
            kendall (bool, optional): Defaults to False.
            missing_feature_percentiles (bool, optional): Defaults to False.
            missing_feature_distribution (bool, optional): Defaults to False.
            median_correlation_outlier_detection (bool, optional): Defaults to False.
            missing_feature_outlier_detection (bool, optional): Defaults to False.
            intensity_analysis (bool, optional): Defaults to False.
            feature_distribution (bool, optional): Defaults to False.
            feature_outlier_detection (bool, optional): Defaults to False.
        
        Args:
            params (dict): the params from the main process.

        Returns:
            list: with all qcqa results for the performed QCQA steps
        """
        self.generate_figure_params(params)
        qaqc_result = []
        for name, method in self.method_map.items():
            if (name in params and params[name]) or ("all" in params and params["all"]):
                try:
                    result = method()
                    if isinstance(result, list):
                        qaqc_result.extend(result)
                    else:
                        qaqc_result.append(result)
                except RecursionError:
                    sys.setrecursionlimit(100000)
                    try:
                        result = method()
                        if isinstance(result, list):
                            qaqc_result.extend(result)
                        else:
                            qaqc_result.append(result)
                    except Exception as e:
                        print("Failure Executing Method: " + name)
                        print(e)
                except Exception as e:
                    print("Failure Executing Method: " + name)
                    print(e)
        return qaqc_result



from mass2chem.formula import compute_adducts_formulae, \
                                    parse_chemformula_dict, \
                                        add_formula_dict, \
                                            dict_to_hill_formula

from .search import build_centurion_tree, find_best_match_centurion_indexed_list


def get_formula_ions_tree(list_formula_mass, mode='pos'):
    peak_list = []
    for formula, mass in list_formula_mass:
        for ion in compute_adducts_formulae(mass, formula, mode, primary_only=True):
            peak_list.append( {'mz': ion[0], 
                               'neutral_formula': formula, 
                               'ion_relation': ion[1],
                               'neutral_formula_mass': mass} )

    return build_centurion_tree(peak_list)

def search_mz_formula_tree(mz, formula_tree, limit_ppm=5):
    '''
    return the best matched ion (as peak format) in formula_tree.
    formula_tree = get_formula_ions_tree(list_formula_mass, mode='pos')
    '''
    return find_best_match_centurion_indexed_list(mz, formula_tree, limit_ppm)

def adjust_charge_in_formula(charged_formula:str,charge:int)->str: 
    """adjust charged formula to neutral formula

    Parameters
    ----------
    charged_formula : str
        charged formula
    charge : int
        charge

    Returns
    -------
    str
        neutral formula

    Examples
    --------
    >>> adjust_charge_in_formula('C17H31O2',-1)
    'C17H32O2'
    
    """ 
    if charge == 0:
        result = charged_formula
    else:
        formula_dict = parse_chemformula_dict(charged_formula)
        hydrogen_dict = {'H':-charge}
        hypothetical_neutral_formula_dict = add_formula_dict(formula_dict, hydrogen_dict)
        if hypothetical_neutral_formula_dict:
            result = dict_to_hill_formula(hypothetical_neutral_formula_dict)
        else: 
            if charge < 0:
                result = None
            else:  # if charge >0, e.g., Fe, charge is 2+; it should still be Fe as formula
                result = charged_formula # here dealing with Metal etc.
    return result



'''
Following tools in mass2chem. 
Some are replicated here to enable standalone use.

tree = build_centurion_tree(list_peaks)

find_all_matches_centurion_indexed_list(112.0950, tree, limit_ppm=5)
Out[17]:
[{'id_number': 'F241',
  'mz': 112.09503429766107,
  'apex': '267',
  'height': '3217405.650383509',
  'cSelectivity': '0.6596858638743456',
  'goodness_fitting': '0.983822971450631',
  'snr': '4'},
 {'id_number': 'F242',
  'mz': 112.09503429766107,
  'apex': '408',
  'height': '2461923.7027675146',
  'cSelectivity': '0.1015625',
  'goodness_fitting': '0.7678341743874929',
  'snr': '3'}]
'''

def build_centurion_tree(list_peaks):
    '''
    list_peaks: [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
                    'left_base': 648, 'right_base': 655, 'id_number': 555}, ...]
    Return a dictionary, indexing mzList by 100*mz bins.
    Because most high-resolution mass spectrometers measure well under 0.01 amu, 
    one only needs to search the corresponding 0.01 bin and two adjacent bins (to capture bordering values).
    list_mass_tracks has similar format as list_peaks.
    '''
    d = {}
    for p in list_peaks:
        cent = int(100 * p['mz'])
        if cent in d:
            d[cent].append(p)
        else:
            d[cent] = [p]
    return d


def build_peak_id_dict(list_peaks):
    d = {}
    for p in list_peaks:
        d[p['id_number']] = p
    return d


def __build_centurion_tree_mzlist(mzList):
    '''
    Return a dictionary, indexing mzList by 100*mz bins.
    Because most high-resolution mass spectrometers measure well under 0.01 amu, 
    one only needs to search the corresponding 0.01 bin and two adjacent bins (to capture bordering values).
    '''
    d = {}
    for ii in range(len(mzList)):
        cent = int(100*mzList[ii])
        if cent in d:
            d[cent].append((mzList[ii], ii))
        else:
            d[cent] = [(mzList[ii], ii)]
    return d


def find_all_matches_centurion_indexed_list(query_mz, mz_centurion_tree, limit_ppm=5):
    '''
    Return matched peaks in mz_centurion_tree.
    '''
    q = int(query_mz * 100)
    mz_tol = query_mz * limit_ppm * 0.000001
    results = []
    for ii in (q-1, q, q+1):
        L = mz_centurion_tree.get(ii, [])
        for peak in L:
            if abs(peak['mz']-query_mz) < mz_tol:
                results.append(peak)
    return results


def find_best_match_centurion_indexed_list(query_mz, mz_centurion_tree, limit_ppm=2):
    '''
    Return matched indices in mz_centurion_tree (based on peak list).
    '''
    q = int(query_mz * 100)
    mz_tol = query_mz * limit_ppm * 0.000001
    result = (None, 999)
    for ii in (q-1, q, q+1):
        L = mz_centurion_tree.get(ii, [])
        for peak in L:
            _d = abs(peak['mz']-query_mz)
            if _d < min(result[1], mz_tol):     # enforce mz_tol here
                result = (peak, _d)
                
    return result[0]


def is_coeluted(P1, P2, rt_tolerance=10):
    '''
    coelution is defined by overlap more than half of the smaller peak, or apexes within rt_tolerance.
    
    Example peak format: {'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0,
    'left_base': 648, 'right_base': 655, 'id_number': 555}
    rt_tolerance: tolerance of retention time is used when peak boundaries are not given.
                Can be scan numbers or seconds.
                
    return: True or False
    '''
    _coeluted = False
    try:
        len1, len2 = P1['right_base'] - P1['left_base'], P2['right_base'] - P2['left_base']
        # overlap is the max L and min R
        overlap = min(P1['right_base'], P2['right_base']) - max(P1['left_base'], P2['left_base'])
        if overlap > 0.5 * min(len1, len2):
            _coeluted = True
            
    except KeyError:
        if abs(P1['apex']-P2['apex']) <= rt_tolerance:
            _coeluted = True
        
    return _coeluted

   
def find_isotopic_signatures(list_peaks, mztree, isotopic_patterns, mz_tolerance_ppm=5, rt_tolerance_scans=5):
    '''
    See find_isotopic_pairs. This extends to all related isotopic signatures.
    Allows ambiguous matches, which are dealt with in empCpd statistics.
    Input
    =====
    isotopic_patterns = [(1.003355, '13C/12C', (0, 0.2)), ...], the third item is optional limits of abundance ratio.
    Return
    ======
    list of lists of peak numbers that match isotopic patterns, [ [], ... ]
    Example
    =======
    [ [(195, 'anchor'), (206, '13C/12C')], 
      [(182, 'anchor'), (191, '13C/12C'), (205, '18O/16O')],
      [(295, 'anchor'), (335, '13C/12C'), (368, 'M(13C),M(34S)')], ...]
    '''

    signatures = []
    for P1 in list_peaks:
        matched = [ (P1['id_number'], 'anchor'), ]          # Nature is nice to have lowest mass for the most abundant 
        for isotopic_pair in isotopic_patterns:
            (mass_difference, relation) = isotopic_pair[:2]
            tmp = find_all_matches_centurion_indexed_list(P1['mz'] + mass_difference, mztree, mz_tolerance_ppm)
            for P2 in tmp:
                if abs(P1['apex']-P2['apex']) <= rt_tolerance_scans and is_coeluted(P1, P2):
                    if len(isotopic_pair) == 2:             # not checking abundance ratio
                        matched.append( (P2['id_number'], relation) )
                    else:
                        (abundance_ratio_min, abundance_ratio_max) = isotopic_pair[2]
                        # checking abundance ratio
                        if abundance_ratio_min*P1['height'] < P2['height'] < abundance_ratio_max*P1['height']:
                            matched.append( (P2['id_number'], relation) )
        if len(matched) > 1:
            signatures.append(matched)

    return signatures


def find_adduct_signatures(list_peaks, mztree, adduct_patterns, mz_tolerance_ppm=5):
    '''
    Search adduct mass_diff in ceelution peaks, de novo. 
    Not requiring matched apex as in isotopic pairs, nor abundance ratio restriction.
    '''
    signatures = []
    for P1 in list_peaks:
        matched = [ (P1['id_number'], 'anchor'), ]
        for adduct in adduct_patterns:
            tmp = find_all_matches_centurion_indexed_list( P1['mz'] + adduct[0], mztree, mz_tolerance_ppm )
            for P2 in tmp:
                if is_coeluted(P1, P2):
                    matched.append( (P2['id_number'], adduct[1]) )

        if len(matched) > 1:
            signatures.append(matched)

    return signatures


def search_formulae(list_peaks, db_tree):
    '''
    Search HMDB and PubChemLite for formula matches
    '''


    pass



def find_mzdiff_pairs_from_masstracks(list_mass_tracks, list_mz_diff=[1.003355, 21.9820], mz_tolerance_ppm=5):
    '''
    Find all pairs in list_mass_tracks that match a pattern in list_mz_diff, and return their id_numbers as pairs.
    This function does not use coeluction (rtime) rules. 
    Input
    =====
    list_mass_tracks: [{ 'id_number': ii,  'mz': xic[0], 'rt_scan_numbers': xic[1],  'intensity': xic[2],  }, ...]
    list_mz_diff: defaul 1.003355, 21.9820 are 13C/12C, Na/H. Dependent on ionization mode.
    
    Return
    ======
    list of pairs of mass tracks numbers.
    '''
    pairs = []
    # list_mass_tracks has similar format as list_peaks.
    mztree = build_centurion_tree(list_mass_tracks)
    for mzdiff in list_mz_diff:
        for P1 in list_mass_tracks:
            P2 = find_best_match_centurion_indexed_list(P1['mz'] + mzdiff, mztree, mz_tolerance_ppm)
            if P2:
                pairs.append((P1['id_number'], P2['id_number']))

    return pairs


def score_emp_cpd_matches(L1_peaks, L2_peaks, ppm):
    '''Return number of matched peaks
    '''
    return len(mass_paired_mapping(L1_peaks, L2_peaks, ppm)[0])


#------------------------------------------------------------------------------------------------------------------------------
# From asari.mass_functions

def mass_paired_mapping(list1, list2, std_ppm=5):
    '''
    To find unambiguous matches of m/z values between two lists.
    This sorts all m/z values first, then compare their differences in sequential neighbors.
    To be considered as an unambiguous match, the m/z values from two lists 
    should have no overlap neighbors in either direction in either list other than their own pair.
    Not necessary to have full mapping. FeatureMap is done in multiple steps. This is step 1.
    This shares some similarity to the RANSAC algorithm but prioritizes selectivity.
    For illustration, one can use one-step Gaussian model for mass shift.
    Since only mean shift is used here, and stdev is implicitly enforced in matching, no need to do model fitting.
    Input
    =====
    Two lists of m/z values, not ncessarily same length.
    std_ppm: instrument accuracy to guide value matching. Low-selectiviy values are not considered in matching.
    Return
    ======
    mapped: mapping list [(index from list1, index from list2), ...]
    ratio_deltas: mean m/z ratio shift between two lists. This is ppm*10^-6. No need to convert btw ppm here.
    Test
    ====
    list1 = [101.0596, 101.061, 101.0708, 101.0708, 101.1072, 101.1072, 101.1072, 102.0337, 102.0337, 102.0548, 102.0661, 102.0912, 102.0912, 102.1276, 102.1276, 103.0501, 103.0501, 103.0541, 103.0865, 103.0865, 103.9554, 104.0368, 104.0705, 104.0705, 104.1069, 104.1069, 104.9922, 105.0422, 105.0698, 105.0698, 105.0738, 105.1039, 105.1102, 105.9955, 106.0497, 106.065, 106.065, 106.0683, 106.0683, 106.0861, 106.0861, 106.0861, 106.1111, 106.9964, 107.0475, 107.0602, 107.0653, 107.0895, 107.9667, 108.0443, 108.0555, 108.0807, 109.0632, 109.0759]
    list2 = [101.0087, 101.035, 101.0601, 101.0601, 101.0601, 101.0601, 101.0713, 101.0714, 101.1077, 101.1077, 101.1077, 101.1077, 101.1077, 101.1158, 101.1158, 102.0286, 102.0376, 102.0468, 102.0539, 102.0554, 102.0554, 102.0554, 102.0554, 102.0666, 102.0917, 102.0917, 102.0917, 102.0918, 102.1281, 102.1281, 102.1282, 103.0394, 103.0505, 103.0507, 103.0547, 103.1233, 103.8162, 103.956, 103.956, 103.956, 104.0532, 104.0533, 104.0641, 104.0709, 104.071, 104.0831, 104.0878, 104.0895, 104.0953, 104.1073, 104.1073, 104.1074, 104.1074, 104.1182, 104.1199, 104.1265, 104.1318, 104.1354, 104.1725, 104.3998, 104.9927, 104.9927, 104.9927, 104.9927, 105.0654, 105.0703, 105.1043, 105.1133, 106.049, 106.0503, 106.0655, 106.0688, 106.0866, 106.0867, 106.0867, 106.0867, 106.114, 107.048, 107.0481, 107.0496, 107.0608, 107.0658, 108.0109, 108.0482, 108.0604, 108.0812, 108.0812, 108.9618, 109.0507, 109.0637, 109.0637, 109.0764, 109.1015]
    mass_paired_mapping(list1, list2) >>>
        ([(10, 23), (29, 65), (31, 66), (36, 70), (38, 71), (46, 81), (53, 91)],
        [4.898762180656323e-06,
        4.758718686464085e-06,
        3.805743437700149e-06,
        4.714068193732999e-06,
        4.713921530199148e-06,
        4.670025348919892e-06,
        4.583942997773922e-06])
    '''
    all = [(list1[ii], 1, ii) for ii in range(len(list1))] + [(list2[jj], 2, jj) for jj in range(len(list2))]
    # [(mz, list_origin, index_origin), ...]
    all.sort()
    NN = len(all)
    # Add a mock entry to allow loop goes through NN.
    all.append((999999, 2, None))
    mapped, ratio_deltas = [], []
    for ii in range(1, NN):
        if all[ii][1] != all[ii-1][1]:          # from two diff list_origin
            _tolerance = all[ii][0] * std_ppm * 0.000001
            _d = all[ii][0]-all[ii-1][0]
            if _d < _tolerance and all[ii+1][0]-all[ii][0] > _tolerance:
                # not allowing ii to be matched to both ii-1 and ii+1
                if all[ii][1] > all[ii-1][1]:   # always ordered as list1, list2
                    mapped.append( (all[ii-1][2], all[ii][2]) )
                    ratio_deltas.append( _d/all[ii][0] )
                else:
                    mapped.append( (all[ii][2], all[ii-1][2]) )
                    ratio_deltas.append( -_d/all[ii][0] )

    return mapped, ratio_deltas



"""
Default parameters for processing

This needs to be cleaned up.
"""

import os

PARAMETERS = {
    #"preprocessing_config": "preprocessing_examples/defaultpreprocessing.json",
    "annot_mz_tolerance": 5,
    "annot_rt_tolerance": 30,
    "khipu_mz_tolerance": 5,
    "khipu_rt_tolerance": 2,
    "mode": None,
    "MS1_annotation_field": "MS1_annotations",
    "MS2_annotation_field": "MS2_annotations",
    "khipu_adducts_pos": "default_configs/default_adducts_pos_khipu.json",
    "khipu_adducts_neg": "default_configs/default_adducts_pos_khipu.json",
    "feature_adducts_pos": "default_configs/default_adducts_pos_feature.json",
    "feature_adducts_neg": "default_configs/default_adducts_neg_feature.json",
    "experiment_config": "default_configs/default_experiment_params.json",
    "khipu_extended_adducts": "default_configs/default_extended_adducts.json",
    "khipu_isotopes": "default_configs/default_isotopes.json",
    "khipu_charges": "default_configs/default_charges.json",
    "auto_drop": "default_configs/default_auto_drop.json",
    "report_config": "default_configs/default_report.json",
    "moniker": "default",
    "multicores": 4,
    "conversion_command": ['$(which mono)', 
                           '$BUILTIN_CONVERTER', 
                           '-f=2', 
                           '-i', 
                           '$RAW_PATH', 
                           '-b',
                           '$OUT_PATH', 
                           ' >/dev/null'],
    "asari_command": [
        'asari',
        'process',
        '-m',
        '$IONIZATION_MODE', 
        '-i',
        '$CONVERTED_SUBDIR',
        '-o',
        '$ASARI_SUBDIR',
    ],
    "filter": '',
    "save_plots": True,
    "interactive_plots": False,
    "color_by": [],
    "marker_by": [],
    "text_by": [],
    "pca": False,
    "tsne": False,
    "pearson": False,
    "kendall": False,
    "spearman": False,
    "missing_feature_percentiles": False,
    "missing_feature_distribution": False,
    "missing_feature_outlier_detection": False,
    "median_correlation_outlier_detection": False,
    "intensity_analysis": False,
    "feature_distribution": False,
    "feature_outlier_detection": False,
    "all": True,
    "blank_value": "blank",
    "sample_value": "unknown",
    "query_field": "Sample Type",
    "blank_intensity_ratio": 3,
    "by_batch": None,
    "batch_blanking_logic": 'or',
    "drop_name": None,
    "drop_field": None,
    "drop_others": None,
    "qaqc_filter": None,
    "drop_value": None,
    "add_singletons": False,
    "TIC_normalization_percentile": 0.90,
    "normalize_value": "median",
    "feature_retention_percentile": 0.50,
    "feature_drop_logic": 'or',
    "interpolation_ratio": 0.5,
    "interpolate_method": "min",
    "log_transform_mode": "log2",
    "targets": ["annotation_sources/hmdb_metabolites.json", "annotation_sources/LMSD.json"],
    "search_isotopologues": False,
    "MS1_annotation_name": "MS1_annotations",
    "MS2_annotation_name": "MS2_annotations",
    "msp_files_pos": "annotation_sources/MoNA-export-LC-MS-MS_Positive_Mode.msp",
    "msp_files_neg": "annotation_sources/MoNA-export-LC-MS-MS_Negative_Mode.msp",
    "ms2_dir": None,
    "ms2_similarity_metric": "CosineGreedy",
    "ms2_min_peaks": 1,
    "find_experiment_ms2": True,
    "requirements_txt": "../requirements.txt",
    "sample_for_ratio": None,
    "deriv_formula": None,
    "skip_list": None,
    "extra_asari": None,
    "seed": None,
    "accept_licenses": False,
    "scan_experiment": True,
    "force": False,
    "file_mode": "link"
}

this_abs_dir = os.path.abspath(os.path.dirname(__file__))

new_target_files = []
for v in PARAMETERS['targets']:
    new_target_files.append(os.path.join(this_abs_dir, v))
PARAMETERS['targets'] = new_target_files

if PARAMETERS["conversion_command"][1] == '$BUILTIN_CONVERTER':
    path = this_abs_dir + '/ThermoRawFileParser.exe'
    
    # leave for now for testing
    #print(os.path.exists(path))
    #os.system("mono " + path + " --version")
    #exit()

    PARAMETERS["conversion_command"][1] = path

for k, v in PARAMETERS.items():
    if isinstance(v, str) and (v.endswith(".json") or v.endswith(".msp") or v.endswith('.txt')):
        PARAMETERS[k] = os.path.join(this_abs_dir, v)
        



'''
# parse HMDB ver 4, single XML files for all metabolites
# Retrieved from https://hmdb.ca/downloads
# Using Python 3 and lxml
# lxml is not in current Python 3 package. 
# It has similar API as ElementTree but faster. To install:
# pip install lxml

In [1]: # cut HMDB metabolite file into 5 pieces                                               

In [2]: infile = 'xml/hmdb_metabolites.xml'                                                    

In [3]: w = open(infile).read()                                                                

In [4]: len(w)                                                                                 
Out[4]: 4077405956

In [5]: len(w)/5                                                                               
Out[5]: 815481191.2

In [6]: w[-30:]                                                                                
Out[6]: 'ations>\n</metabolite>\n</hmdb>\n'

In [7]: w[:30]                                                                                 
Out[7]: '<?xml version="1.0" encoding="'

In [8]: w[:300]                                                                                
Out[8]: '<?xml version="1.0" encoding="UTF-8"?>\n<hmdb xmlns="http://www.hmdb.ca">\n<metabolite>\n  <version>4.0</version>\n  <creation_date>2005-11-16 15:48:42 UTC</creation_date>\n  <update_date>2020-04-23 20:53:36 UTC</update_date>\n  <accession>HMDB0000001</accession>\n  <status>quantified</status>\n  <secondary'

In [9]: w[-300:]                                                                               
Out[9]: 'ctrometry for metabolomic studies of human urine. Anal Chem. 2012 Feb 21;84(4):1994-2001. doi: 10.1021/ac2030738. Epub 2012 Feb 1.</reference_text>\n      <pubmed_id>22409530</pubmed_id>\n    </reference>\n  </general_references>\n  <protein_associations>\n  </protein_associations>\n</metabolite>\n</hmdb>\n'

In [10]: block = 815481191                                                                     

In [11]:                                                                                       

In [11]: w[:300].find('<metabolite>')                                                          
Out[11]: 73

In [12]: w[:300][73]                                                                           
Out[12]: '<'

In [13]: w[:300][72]                                                                           
Out[13]: '\n'

In [14]: w[block:].find('<metabolite>')                                                        
Out[14]: 14578

In [15]: part1 = w[: block+14578]                                                              

In [16]: part1[-11:]                                                                           
Out[16]: 'etabolite>\n'

In [17]: part1[-100:]                                                                          
Out[17]: 'eferences>\n  </general_references>\n  <protein_associations>\n  </protein_associations>\n</metabolite>\n'

In [18]: tail = '</hmdb>\n'                                                                    

In [19]: with open('HMDB4_part1.xml', 'w') as O: 
    ...:     O.write(part1 + tail) 
    ...:                                                                                       

In [20]: del part1                                                                             

In [21]: w = w[block+14578:]                                                                   

In [22]:                                                                                       

In [22]: w[:80]                                                                                
Out[22]: '<metabolite>\n  <version>4.0</version>\n  <creation_date>2012-09-13 11:52:34 UTC</'

In [23]: w[block:].find('<metabolite>')                                                        
Out[23]: 7241

In [24]: part2 = w[: block+7241]                                                               

In [25]: w = w[block+7241:]                                                                    

In [26]: with open('HMDB4_part2.xml', 'w') as O: 
    ...:     O.write(part2+tail) 
    ...:                                                                                       

In [27]: del part2                                                                             

In [28]:                                                                                       

In [28]: w[block:].find('<metabolite>')                                                        
Out[28]: 4790

In [29]: head = '<?xml version="1.0" encoding="UTF-8"?>\n<hmdb xmlns="http://www.hmdb.ca">\n'  

In [30]: part3 = w[:block+4790]                                                                

In [31]: w = w[block+4790:]                                                                    

In [32]: with open('HMDB4_part3.xml', 'w') as O: 
    ...:     O.write( head + part3 + tail ) 
    ...:                                                                                       

In [33]: del part3                                                                             

In [34]:                                                                                       

In [34]: w[block:].find('<metabolite>')                                                        
Out[34]: 11469

In [35]: part4 = w[:block+11469]                                                               

In [36]: w = w[block+11469:]                                                                   

In [37]:                                                                                       

In [37]: with open('HMDB4_part4.xml', 'w') as O: 
    ...:     O.write( head + part4 + tail ) 
    ...:                                                                                       

In [38]: del part4                                                                             

In [39]:                                                                                       

In [39]: with open('HMDB4_part5.xml', 'w') as O: 
    ...:     O.write( head + w ) 
    ...:                                                                                       

In [40]: del w                                                                                 

In [41]:                                                                                       

In [41]: # fix part2                                                                           

In [42]: part2 = open('HMDB4_part2.xml').read()                                                

In [43]: with open('HMDB4_part2.xml', 'w') as O: 
    ...:     O.write( head + part2 ) 
    ...:                                                                                       

In [44]: del part2                                                                             

In [45]:     

# parsing using this script

# This took ~5 seconds on my i5 laptop, infile is 1.2 GB.
# Larger files will have to be partitioned, depending on memeory limit.
In [1]: from parse_hmdb4 import *                                                              
In [2]: myList = parse_a_file(infile, wanted)                                                  
Found 25411 entries.
In [3]: myList[0]                                                                              
Out[3]: 
{'accession': 'HMDB0000001',
 'name': '1-Methylhistidine',
 'chemical_formula': 'C7H11N3O2',
 'monisotopic_moleculate_weight': '',
 'iupac_name': '(2S)-2-amino-3-(1-methyl-1H-imidazol-4-yl)propanoic acid',
 'traditional_iupac': '1 methylhistidine',
 'cas_registry_number': '332-80-9',
 'smiles': 'CN1C=NC(C[C@H](N)C(O)=O)=C1',
 'inchi': 'InChI=1S/C7H11N3O2/c1-10-3-5(9-4-10)2-6(8)7(11)12/h3-4,6H,2,8H2,1H3,(H,11,12)/t6-/m0/s1',
 'inchikey': 'BRMWTNUJHUMWMS-LURJTMIESA-N',
 'pathways': '',
 'normal_concentrations': '',
 'abnormal_concentrations': '',
 'diseases': '',
 'drugbank_id': 'DB04151',
 'drugbank_metabolite_id': '',
 'phenol_explorer_compound_id': '',
 'phenol_explorer_metabolite_id': '',
 'foodb_id': 'FDB093588',
 'knapsack_id': '',
 'chemspider_id': '83153',
 'kegg_id': 'C01152',
 'biocyc_id': '',
 'bigg_id': '',
 'wikipidia': '',
 'nugowiki': '',
 'metagene': '',
 'metlin_id': '3741',
 'pubchem_compound_id': '92105',
 'het_id': '',
 'chebi_id': '50599',
 'protein_associations': ''}
In [40]: with open('HMDB0000031.xml', 'w') as O: 
    ...:     O.write(str(ET.tostring(x), 'utf-8')) 
    ...:                                                                                       
# Alternatively, cut the big file into 5 parts, then parse:
In [2]: myResults = []                                                                         
In [3]: wanted = ['accession', 'name', 'chemical_formula', 'monisotopic_molecular_weight', 'keg
   ...: g_id', 'pubchem_compound_id', ]                                                        
In [4]: import os                                                                              
In [5]: files = os.listdir('parts/')                                                           
In [6]: files                                                                                  
Out[6]: 
['README.txt',
 'HMDB4_part5.xml',
 'HMDB4_part3.xml',
 'HMDB4_part2.xml',
 'HMDB4_part1.xml',
 'HMDB4_part4.xml']
In [7]: for f in files[1:]: 
   ...:     myResults += parse_a_file('parts/' + f, wanted) 
   ...:                                                                                        
Found 30770 entries.
Found 21801 entries.
Found 17478 entries.
Found 22506 entries.
Found 21667 entries.
In [8]: write_tsv(myResults, wanted, "HMDB4_compounds.tsv")                
'''

from lxml import etree as ET 

infile = 'serum_metabolites.xml'

# fields to retrieve
wanted = ['accession', 'name', 'chemical_formula', 'monisotopic_molecular_weight', 'iupac_name', 
    'traditional_iupac', 'cas_registry_number', 'smiles', 'inchi', 'inchikey', 'pathways', 
    'normal_concentrations', 'abnormal_concentrations', 'diseases', 'drugbank_id', 
    'drugbank_metabolite_id', 'phenol_explorer_compound_id', 'phenol_explorer_metabolite_id', 
    'foodb_id', 'knapsack_id', 'chemspider_id', 'kegg_id', 'biocyc_id', 'bigg_id', 'wikipidia', 
    'nugowiki', 'metagene', 'metlin_id', 'pubchem_compound_id', 'het_id', 'chebi_id', 'protein_associations']

def extract(obj_metabolite, x):
    try:
        y = obj_metabolite.find(x).text.strip()
        if y:
            return y
        else: return ''
    except AttributeError: 
        return ''

def extract_dict(obj_metabolite, wanted, prefix):
    result = {}
    for x in wanted:
        result[x] = extract(obj_metabolite, prefix+x)
    return result

def parse_a_file(f, wanted, prefix='{http://www.hmdb.ca}'):
    '''
    Input
    =====
    f: HMDB 4 XML file
    wanted: fields to retrieve
    prefix: prefix that is used in the XML file
    Return
    ======
    return as a list of dictionaries. 
    '''
    tree = ET.parse(f)
    root = tree.getroot()
    print("Found %d entries." %len(root))

    results = []
    for child in root:
        results.append(extract_dict(child, wanted, prefix=prefix))

    return results

def write_tsv(results, wanted, outfile):
    s = '\t'.join(wanted) + '\n'
    for R in results:
        s += '\t'.join([R[x] for x in wanted]) + '\n'
    with open(outfile, 'w') as O:
        O.write(s)

if __name__ == '__main__':
    # wanted = ['accession', 'name', 'chemical_formula', 'monisotopic_molecular_weight']
    infile = 'serum_metabolites.xml'
    # infile= 'urine_metabolites.xml'
    # infile= 'feces_metabolites.xml'

    write_tsv(
        parse_a_file(infile, wanted),
        wanted, 
        "parsed_"+infile.replace(".xml", ".tsv")
    )



'''
# decompress the HMDB4 data if needed
➜  JMS git:(main) ✗ xz -d jms/data/compounds/list_compounds_HMDB4.json.xz 
# run test from top dir
➜  JMS git:(main) ✗ python3 -m jms.test 
'''

import json
from jms.dbStructures import annotate_peaks_against_kcds
from jms.io import read_table_to_peaks

list_compounds = json.load(open('jms/data/compounds/list_compounds_HMDB4.json'))
mydata = read_table_to_peaks('testdata/full_Feature_table.tsv', '\t')

annotate_peaks_against_kcds(mydata, list_compounds, 
                                export_file_name_prefix='jms_annotated_',
                                mode='pos',  mz_tolerance_ppm=5)




'''
Conversion of genome scale metabolic models to JSON formats,and to indexed Python internal objects.
Matching btw metabolic models and experimental empirical compounds.
Used in mummichog 3.
'''

from khipu.epdsConstructor import epdsConstructor
from .dbStructures import knownCompoundDatabase, ExperimentalEcpdDatabase

default_parameters = {'mode': 'pos',
                        'mz_tolerance_ppm': 5,
                        'rt_tolerance': 2,
                        'isotope_search_patterns': [(1.003355, '13C/12C', (0, 0.8)),
                                (2.00671, '13C/12C*2', (0, 0.8))],
                        'adduct_patterns': [(21.982, 'Na/H'),
                                (41.026549, 'ACN'),
                                (35.9767, 'HCl'),
                                (37.955882, 'K/H')],
                        'extended_adducts': [(1.0078, 'H'),
                                (-1.0078, '-H'),
                                (10.991, 'Na/H, double charged'),
                                (0.5017, '13C/12C, double charged'),
                                (117.02655, '-NH3'),
                                (17.02655, 'NH3'),
                                (-18.0106, '-H2O'),
                                (18.0106, 'H2O'),
                                (18.033823, 'NH4'),
                                (27.01089904, 'HCN'),
                                (37.94694, 'Ca/H2'),
                                (32.026215, 'MeOH'),
                                (43.96389, 'Na2/H2'),
                                (67.987424, 'NaCOOH'),
                                (83.961361, 'KCOOH'),
                                (97.96737927, 'H2SO4'),
                                (97.97689507, 'H3PO4')],
 }

def convert_json_model(jmodel):
    '''
    Convert a JSON style, concise genome scale metabolic models to mummichog style with internal indexing.
    All compound identifiers and other identifiers are expected to be contained within the model.
    Returns model as a dictionary with as in mummichog style.

    >>> jmodel.keys()
        dict_keys(['id', 'list_of_reactions', 'list_of_compounds', 'list_of_pathways', 'meta_data'])
    >>> jmodel['list_of_compounds'][772]
    {'id': 'ebastineoh', 'name': 'Hydroxylated Ebastine', 'identifiers': {'inchi': 'InChI=1S/C32H39NO3/c1-32(2,24-34)28-17-15-25(16-18-28)30(35)14-9-21-33-22-19-29(20-23-33)36-31(26-10-5-3-6-11-26)27-12-7-4-8-13-27/h3-8,10-13,15-18,29,31,34H,9,14,19-24H2,1-2H3'}, 'neutral_formula': '', 'charge': 0, 'charged_formula': 'C32H39NO3', 'neutral_mono_mass': 0.0, 'SMILES': '', 'inchi': ''}
    >>> jmodel['list_of_pathways'][2]
    {'id': 'group3', 'name': 'Aminoacyl-tRNA biosynthesis', 
    'list_of_reactions': ['HMR_5130', 'HMR_5131', 'HMR_5132', 'HMR_5133', 'HMR_5134', 'HMR_5135', 
            'HMR_5136', 'HMR_5137', 'HMR_5138', 'HMR_5139', 'HMR_5140', 'HMR_5141', 'HMR_5142', 'HMR_5143', 
            'HMR_5144', 'HMR_5145', 'HMR_5146', 'HMR_5147', 'HMR_5148', 'HMR_5149', 'HMR_5150']}
    >>> 
    >>> jmodel['list_of_reactions'][22]
    {'id': '24_25DHVITD3tm', 'reactants': ['2425dhvitd3'], 'products': ['2425dhvitd3']}
    >>> mfn['metabolic_pathways'][51]
    {'cpds': ['C00024', 'C00027', 'strdnccoa', 'C00100', 'C01944', 'q10h2', 'dmhptcoa', 'etfrd', 'C03035', 'etfox', 'dmnoncoa', 'arachdcoa', 'dcsptn1coa', 'ptdcacoa', 'C00412', 'od2coa', 'hdcoa', 'C02593', 'dlnlcgcoa', 'vacccoa', 'C00010', 'C00016', 'adrncoa', 'C00154', 'lneldccoa', 'lnlccoa', 'c226coa', 'lnlncacoa', 'C00399', 'odecoa', 'hpdcacoa', 'C01352', 'eicostetcoa', 'tmndnccoa', 'arachcoa'], 
    'rxns': ['FAOXC2242046m', 'FAOXC200180m', 'FAOXC204', 'FAOXC160', 'FAOXC16080m', 'FAOXC180', 'FAOXC80', 'FAOXC140', 'FAOXC1811601m', 'FAOXC1811603m', 'FAOXC150m', 'FAOXC18480m', 'FAOXC1811602m', 'FAOXC16180m', 'FAOXC170m', 'FAOXC18280m', 'FAOXC182806m', 'FAOXC183803m', 'FAOXC2252053m', 'FAOXC2031836m', 'FAOXC11', 'FAOXC204184m', 'FAOXC226205m', 'FAOXC226', 'ETFQO', 'ETF'], 
    'ecs': ['1.3.99.3', '1.5.5.1'], 'id': 'mfn1v10path136', 'name': 'Fatty acid oxidation'}
    >>> mfn['metabolic_rxns'][2]
    {'reactants': ['C06611'], 'id': 'R05233', 'source': 'kegg_dre', 'ecs': ['1.1.1.1'], 'products': ['C06613'], 'cpds': ['C06611', 'C06613'], 'pathway': '3-Chloroacrylic acid degradation'}
    >>>    
    mfn['cpd2pathways'] = {
         'C06125': ['Glycosphingolipid metabolism', 'Glycosphingolipid biosynthesis - ganglioseries'], ...
         }
    >>> mfn['cpd_edges'][:5]
[   ['C00106', 'C00526'], ['C05977', 'CE6326'], ['C00043', 'G00174'], ['C00199', 'C00231'], ['C00014', 'C00655']]
    '''
    new = {}
    new['id'] = jmodel['id']
    new['version'] = jmodel['meta_data']['version']
    cpdDict,  dict_cpds_def = {}, {}
    for cpd in jmodel['list_of_compounds']:
        if 'neutral_mono_mass' in cpd:
            cpd['neutral_formula_mass'] = cpd['neutral_mono_mass']
        if 'neutral_formula_mass' in cpd and cpd['neutral_formula_mass']:
            cpdDict[cpd['id']] = cpd    # convert_compound_azimuth_mummichog(cpd, hmdb_dict, kegg_dict)
            dict_cpds_def[cpd['id']] = cpd['name']
    new['Compounds'] = cpdDict
    new['dict_cpds_def'] = dict_cpds_def
    new['metabolic_rxns'] = jmodel['list_of_reactions']

    # bunch of indexing
    dict_rxns, edge2rxn, edge2enzyme = {}, {}, {}
    cpd_edges = []
    for rxn in jmodel['list_of_reactions']:
        dict_rxns[rxn['id']] = rxn
        _edges = []
        for cpd1 in rxn['reactants']:
            for cpd2 in rxn['products']:
                k = [cpd1, cpd2]
                cpd_edges.append(k)
                _edges.append(k)
                edge2rxn[','.join(sorted(k))] = rxn['id']
                enzymes = rxn.get('ecs', []) or rxn.get('enzymes', [])
                for e in enzymes:
                    edge2enzyme[','.join(sorted(k))] = e

    new['cpd_edges'] = cpd_edges
    new['edge2rxn'] = edge2rxn
    new['edge2enzyme'] = edge2enzyme
    # now do pathways
    metabolic_pathways = []
    for P in jmodel['list_of_pathways']:
        _p = {}
        _p['id'] = P['id']
        _p['name'] = P['name']
        _p['rxns'] = P['list_of_reactions']
        cpds, ecs, genes = [], [], []
        for ii in P['list_of_reactions']:
            rxn = dict_rxns[ii]
            cpds += rxn['reactants'] + rxn['products']
            enzymes = rxn.get('ecs', []) or rxn.get('enzymes', [])
            ecs += enzymes
            if 'genes' in rxn:
                genes += rxn['genes']
        _p['cpds'] = list(set(cpds))
        _p['ecs'] = list(set(ecs))
        _p['genes'] = list(set(genes))
        metabolic_pathways.append(_p)

    cpd2pathways = {}
    for _p in metabolic_pathways:
        for cpd in _p['cpds']:
            if cpd in cpd2pathways:
                cpd2pathways[cpd].append(_p['name'])
            else:
                cpd2pathways[cpd] = [_p['name']]

    new['metabolic_pathways'] = metabolic_pathways
    new['cpd2pathways'] = cpd2pathways

    return new



class metabolicNetwork:
    '''
    Simple wrapper class of a genome scale metabolic model in JSON.

    ?? keep or not ??
    '''
    def __init__(self, MetabolicModel):
        '''
        MetabolicModel : metabolic model in JSON style dictionary.
        '''
        #print_and_loginfo( "Loading metabolic network %s..." %MetabolicModel.version ) # version from metabolic model
        
        self.MetabolicModel = MetabolicModel
        self.network = self.build_network(MetabolicModel['cpd_edges'])
        
        self.version = MetabolicModel['version']
        self.Compounds = MetabolicModel['Compounds']
        self.metabolic_pathways = MetabolicModel['metabolic_pathways']
        self.dict_cpds_def = MetabolicModel['dict_cpds_def']
        self.cpd2pathways = MetabolicModel['cpd2pathways']
        self.edge2enzyme = MetabolicModel['edge2enzyme']
        self.total_cpd_list = self.network.nodes()
        
        
    def build_network(self, edges):
        '''
        import networkx as nx
        '''
        return nx.from_edgelist( edges )
        

    def get_pathways(self):
        pass

    def validate(self):
        pass


class DataMeetModel:
    '''
    Match the compounds in a genome scale metabolic model to a metabolomics experiment,
    via the format of empirical compounds. The empirical compounds are constructed via khipu.
    Their JSON format can accommodate added fields on the fly, and return results 
    as part of empCpd['identity'] (https://github.com/shuzhao-li/metDataModel).
    Because neutral mass is inferred from khipu, the match problem is simplified by 
    focusing on neutral mass (formula).
    '''
    def __init__(self, MetabolicModel, userFeatureList=None, userListEmpCpds=None, parameters=default_parameters):
        '''
        Besides MetabolicModel, this takes either userFeatureList or userListEmpCpds.
        parameters : dictionary to pass ion mode, m/z and rt tolerance and isotope/adduct patterns.
        MetabolicModel : metabolic model in JSON style dictionary.
        userFeatureList : list of JSON style features
        userListEmpCpds : list of empirical compounds, which can be constructed and processed elsewhere
        '''
        self.model = MetabolicModel
        self.userFeatureList = userFeatureList
        self.userListEmpCpds = userListEmpCpds
        
        self.mode = parameters['mode']
        self.isotope_search_patterns = parameters['isotope_search_patterns']
        self.adduct_patterns = parameters['adduct_patterns']
        self.extended_adducts = parameters['extended_adducts']
        self.mz_tolerance_ppm = parameters['mz_tolerance_ppm']
        self.rt_tolerance = parameters['rt_tolerance']

    def match_all(self):
        '''
        Match model compounds with empirical compounds, via the JMS KCD-EED architecture.
        The search here cannot distinguish isomers, thus centering on neutral mass/formula.

        Returns
        -------
        dict_empCpds : {id: empCpd, ...} with matched compounds in empCpd['list_matches'],
            which have KCD empCpd identifiers but compound records need to pull out KCD later.
            This dict includes singletons and empCpds without matches.

        Examples
        --------
        A intermediary EED.dict_empCpds before updating identity: 
        {'interim_id': 'kp203_202.1317', 
            'neutral_formula_mass': 202.13169603323, 'neutral_formula': None, 
            'Database_referred': [], 'identity': [], 
            'MS1_pseudo_Spectra': [{'id_number': 'F3900', 'mz': 204.1424, 'rtime': 23.04, 'rtime_left_base': '21.88', 
                    'rtime_right_base': '25.37', 'parent_masstrack_id': '2519', 'peak_area': '111965925', 'cSelectivity': '0.86', 
                    'goodness_fitting': '0.98', 'snr': '6675', 'detection_counts': '15', ' 'apex': 23.04, 
                    'representative_intensity': '111965925', 'id': 'F3900', 
                    'isotope': '13C/12C', 'modification': 'M+H+', 'ion_relation': '13C/12C,M+H+', 'parent_epd_id': 'kp203_202.1317'}, 
            {'id_number': 'F3684', 'mz': 203.1389, 'rtime': 23.04, 'rtime_left_base': '21.88', 'rtime_right_base': '25.37', 
                    'parent_masstrack_id': '2495', 'peak_area': '402792917', 'cSelectivity': '0.86', 'goodness_fitting': '0.98', 
                    'snr': '120', 'detection_counts': '15',  'apex': 23.04, 'representative_intensity': '402792917', 
                    'id': 'F3684', 'isotope': 'M0', 'modification': 'M+H+', 'ion_relation': 'M0,M+H+', 'parent_epd_id': 'kp203_202.1317'}], 
            'MS2_Spectra': [], 
            'list_matches': [('C9H18N2O3_202.131742', 'neutral', 1)]}

        An indexed KCD empCpd, KCD.mass_indexed_compounds['C9H18N2O3_202.131742']:
        {'interim_id': 'C9H18N2O3_202.131742',
            'neutral_formula': 'C9H18N2O3',
            'neutral_formula_mass': 202.13174244717,
            'compounds': [{'id': 'MAM03375',
            'name': 'L-Alanyl-L-Leucine',
            'identifiers': [['humanGEM', 'MAM03375'],
                ['bigg.metabolite', 'CE5866'],
                ['pubchem.compound', '6992388'],
                ['vmhmetabolite', 'CE5866'],
                ['metanetx.chemical', 'MNXM15786'],
                ['inchi',
                'InChI=1S/C9H18N2O3/c1-5(2)4-7(9(13)14)11-8(12)6(3)10/h5-7H,4,10H2,1-3H3,(H,11,12)(H,13,14)/t6-,7+/m1/s1']],
            'neutral_formula': 'C9H18N2O3',
            'charge': 0,
            'charged_formula': 'C9H18N2O3',
            'neutral_mono_mass': 202.13174244717,
            'SMILES': '',
            'inchi': '',
            'neutral_formula_mass': 202.13174244717}]}
        '''
        KCD = knownCompoundDatabase()
        KCD.mass_index_list_compounds(self.model['Compounds'].values())
        KCD.build_emp_cpds_index()
        EED = ExperimentalEcpdDatabase(mode=self.mode, 
                                       mz_tolerance_ppm=self.mz_tolerance_ppm, 
                                       rt_tolerance=self.rt_tolerance)
        if self.userFeatureList:
            EED.build_from_list_peaks(self.userFeatureList)
        elif self.userListEmpCpds:
            EED.build_from_list_empCpds(self.userListEmpCpds)
        EED.extend_empCpd_annotation(KCD)
        EED.annotate_singleton_mummichog(KCD)

        return self.update_identity(EED.dict_empCpds, KCD)


    def update_identity(self, dict_empCpds, KCD):
        '''
        Updates dict_empCpds by adding field `identity` with compound IDs.
        E.g. {'identity': ['MAM03375'], ... }
        This overwrites field `identity` if it exists.
        '''
        def _get_kcd_ids_(interim_id, KCD):
            e = KCD.mass_indexed_compounds.get(interim_id, None)
            if e:
                return [x['id'] for x in e['compounds']]
            else:
                return []

        for k,v in dict_empCpds.items():
            if 'list_matches' in v:
                v['identity'] = []
                for M in v['list_matches']:
                    v['identity'] += _get_kcd_ids_(M[0], KCD)

        return dict_empCpds


    def _construct_EmpiricalCompounds_(self):
        '''
        For testing only. Use EED class for full application, as in self.match_all().
        Returns
        -------
        dict_empCpds : {id: empCpd, ...}
        '''
        ECCON = epdsConstructor(self.userFeatureList, mode=self.mode)
        return ECCON.peaks_to_epdDict(
                        self.isotope_search_patterns,
                        self.adduct_patterns,
                        self.extended_adducts,
                        self.mz_tolerance_ppm,
                        self.rt_tolerance
        )
